from __future__ import division

import math
import cmath
import numbers
import warnings

try:
    from itertools import izip  # Python 2
except ImportError:
    izip = zip

from GTC import lib
from GTC import reporting
from GTC import type_b
from GTC import type_a
from GTC import persistence
from GTC import function

from GTC import (   
    inf,
    nan,
    is_sequence,
    copyright,
    version
)

# aliases 
rp = reporting
tb = type_b
ta = type_a
fn = function
pr = persistence

value = type_b.value

UncertainReal = lib.UncertainReal
UncertainComplex = lib.UncertainComplex

__all__ = (
        'ureal'
    ,   'multiple_ureal'
    ,   'multiple_ucomplex'
    ,   'ucomplex'
    ,   'constant'
    ,   'value'
    ,   'uncertainty'
    ,   'variance'
    ,   'dof'
    ,   'label'
    ,   'component'
    ,   'inf'
    ,   'nan'
    ,   'get_covariance'
    ,   'get_correlation'
    ,   'set_correlation'
    ,   'result'
    ,   'cos'
    ,   'sin'
    ,   'tan'
    ,   'acos'
    ,   'asin'
    ,   'atan'
    ,   'atan2'
    ,   'exp'
    ,   'pow'
    ,   'log'
    ,   'log10'
    ,   'sqrt'
    ,   'sinh'
    ,   'cosh'
    ,   'tanh'
    ,   'acosh'
    ,   'asinh'
    ,   'atanh'
    ,   'mag_squared'
    ,   'magnitude'
    ,   'phase'
    ,   'copyright'
    ,   'version'
    ,   'reporting',    'rp'
    ,   'function',     'fn'
    ,   'type_b',       'tb'
    ,   'type_a',       'ta'
    ,   'persistence',  'pr'
    ,   'linear_algebra', 'la'
    ,   'math'
    ,   'cmath'
)


# #----------------------------------------------------------------------------
# def value(x):
    # """Return the value 
    
    # Returns a complex number if ``x`` is an uncertain complex number
    
    # Returns a real number if ``x`` is an uncertain real number
    
    # Returns ``x`` otherwise.

    # **Example**::

        # >>> un = ureal(3,1)
        # >>> value(un)
        # 3.0
        # >>> un.x
        # 3.0

    # """
    # try:
        # return x.x
    # except AttributeError:
        # return x
    
#----------------------------------------------------------------------------
def uncertainty(x):
    """
    Return the standard uncertainty

    If ``x`` is an uncertain complex number,
    return a 2-element sequence containing the standard
    uncertainties of the real and imaginary components. 
    
    If ``x`` is an uncertain real number,
    return the standard uncertainty.
    
    Otherwise, return 0.    

    **Examples**::

        >>> ur = ureal(2.5,0.5,3,label='x')
        >>> uncertainty(ur)
        0.5
        >>> ur.u
        0.5
        
        >>> uc = ucomplex(1+2j,(.5,.5),3,label='x')
        >>> uncertainty(uc)
        StandardUncertainty(real=0.5, imag=0.5)
    
    """
    try:
        return x.u
    except AttributeError:
        return 0.0
    
#----------------------------------------------------------------------------
def variance(x):
    """
    Return the standard variance 

    If ``x`` is an uncertain real number, return the
    standard variance.
    
    If ``x`` is an uncertain complex number, return
    a 4-element sequence containing elements of the
    variance-covariance matrix.
    
    Otherwise, return 0.    
    
    **Examples**::

        >>> ur = ureal(2.5,0.5,3,label='x')
        >>> variance(ur)
        0.25
        >>> ur.v
        0.25
        
        >>> uc = ucomplex(1+2j,(.5,.5),3,label='x')
        >>> variance(uc)
        VarianceCovariance(rr=0.25, ri=0.0, ir=0.0, ii=0.25)
    
    """
    try:
        return x.v
    except AttributeError:
        return 0.0
    
#----------------------------------------------------------------------------
def dof(x):
    """
    Return the degrees-of-freedom 

    Returns ``inf`` when the degrees of freedom is greater than 1E6
    
    **Examples**::

        >>> ur = ureal(2.5,0.5,3,label='x')
        >>> dof(ur)
        3.0
        >>> ur.df
        3.0
        
        >>> uc = ucomplex(1+2j,(.3,.2),3,label='x')
        >>> dof(uc)
        3.0
    
    """
    try:
        return x.df
    except AttributeError:
        return inf

#---------------------------------------------------------------------------
def component(y,x):
    """
    Return the magnitude of the component of uncertainty 
    in ``y`` due to ``x``.

    :arg y: an uncertain number
    :type y: :class:`~lib.UncertainReal` or :class:`~lib.UncertainComplex`

    :arg x: an uncertain number
    :type x: :class:`~lib.UncertainReal` or :class:`~lib.UncertainComplex`

    :rtype: float
    
    If ``x`` and ``y`` are uncertain real, the function calls 
    :func:`reporting.u_component` and returns the magnitude 
    of the result. 
    
    If either ``x`` or ``y`` is uncertain complex,
    the returned value represents the magnitude 
    of the component of uncertainty matrix (this is 
    obtained by applying :func:`reporting.u_bar`    
    to the result obtained from :func:`reporting.u_component`).

    If either ``x`` or ``y`` is a number, zero is returned.
    
    ``component`` can also e used in conjunction with :func:`~core.result` 
    to evaluate a component of uncertainty with respect to an 
    intermediate uncertain number. 
    
    **Examples**::
    
        >>> x1 = ureal(2,1)
        >>> x2 = ureal(5,1)
        >>> y = x1/x2
        >>> reporting.u_component(y,x2)
        -0.08
        >>> component(y,x2)
        0.08
        
        >>> z1 = ucomplex(1+2j,1)
        >>> z2 = ucomplex(3-2j,1)
        >>> y = z1 - z2
        >>> reporting.u_component(y,z2)
        ComponentOfUncertainty(rr=-1.0, ri=0.0, ir=0.0, ii=-1.0)
        >>> component(y,z2)
        1.0

        >>> I = ureal(1E-3,1E-5)
        >>> R = ureal(1E3,1)
        >>> V = result( I*R )
        >>> P = V**2/R  
        >>> component(P,V)   
        2.0099751242241783e-05
        
    """
    return reporting.u_bar( y.u_component(x) )
            
#----------------------------------------------------------------------------
def label(x):
    """
    Return the label
    
    """
    return x.label 
    
#----------------------------------------------------------------------------
def ureal(x,u,df=inf,label=None,independent=True):
    """
    Create an elementary uncertain real number

    :arg x: the value (estimate)
    :type x: float
    
    :arg u: the standard uncertainty
    :type u: float
    
    :arg df: the degrees-of-freedom
    :type df: float
    
    :arg label: a string label 
    :type label: str 
    
    :arg independent: not correlated with other UNs
    :type independent: bool
    
    :rtype: :class:`~lib.UncertainReal`
    
    **Example**::
    
        >>> ur = ureal(2.5,0.5,3,label='x')
        >>> ur
        ureal(2.5,0.5,3.0, label='x')
    
    """    
    # Arguments to these math functions must be compatible with float
    if math.isnan(x) or math.isinf(x):
        raise ValueError("invalid: '{!r}'".format(x) )

    if u < 0 or math.isinf(u) or math.isnan(u):
        raise ValueError("invalid uncertainty: '{!r}'".format(u) )

    # inf is allowed, but not nan
    if df < 1 or math.isnan(df):
        raise ValueError("invalid dof: '{!r}'".format(df) )
    
    if u == 0:
        # Is this what we want? Perhaps not.
        return UncertainReal._constant(float(x),label)

    else:
        return UncertainReal._elementary(
            float(x),
            float(u),
            float(df),
            label,
            independent
        )
        
#---------------------------------------------------------------------------
# TODO: think of a better name! Perhaps `ureal_ensemble`
def multiple_ureal(x_seq,u_seq,df,label_seq=None):
    """Return a sequence of related elementary uncertain real numbers

    :arg x_seq: a sequence of values (estimates)
    :arg u_seq: a sequence of standard uncertainties
    :arg df: the degrees-of-freedom 
    :arg label_seq: a sequence of labels
    
    :rtype: a sequence of :class:`~lib.UncertainReal`

    Defines an set of uncertain real numbers with 
    the same number of degrees-of-freedom.
    
    Correlation between any pairs of this set of uncertain  
    numbers defined will not invalidate degrees-of-freedom 
    calculations.
    (see: R Willink, *Metrologia* 44 (2007) 340-349, Sec. 4.1)
    
    **Example**::
    
        # Example from GUM-H2
        >>> x = [4.999,19.661E-3,1.04446]
        >>> u = [3.2E-3,9.5E-6,7.5E-4]
        >>> labels = ['V','I','phi']
        >>> v,i,phi = multiple_ureal(x,u,4,labels)
   
        >>> set_correlation(-0.36,v,i)
        >>> set_correlation(0.86,v,phi)
        >>> set_correlation(-0.65,i,phi)

        >>> r = v/i*cos(phi)
        >>> r
        ureal(127.732169928102...,0.0699787279883717...,4.0)

    """
    if len(x_seq) != len(u_seq):
        raise RuntimeError(
            "unequal length sequences: x={!r} u={!r}".format(x_seq,u_seq)
        )

    if label_seq is None:
        label_seq = [None]*len(x_seq)
    elif is_sequence(label_seq):
        if len(x_seq) != len(label_seq):
            raise RuntimeError(
                "unequal length sequences: x={!r} label_seq={!r}".format(x_seq,u_seq)
            )
    else:
        raise RuntimeError(
            "invalid `label_seq`: {!r}".format(label_seq)
        )
        
    rtn = [
        # NB `ureal` creates constant objects when u == 0
        ureal(x_i,u_i,df,label=l_i,independent=False)
            for x_i,u_i,l_i in izip(
                x_seq,u_seq,label_seq
            )
    ]

    # Only non-constant UNs can be collected in an ensemble
    lib.real_ensemble( 
        [ un_i 
            for un_i in rtn 
                if not lib._is_uncertain_real_constant(un_i) 
        ], df 
    )

    # All uncertain numbers are returned, including the constants
    return rtn

#----------------------------------------------------------------------------
def constant(x,label=None):
    """
    Create a constant uncertain number (with no uncertainty)

    :arg x: a number
    :type x: float or complex

    :rtype: :class:`~lib.UncertainReal` or :class:`~lib.UncertainComplex`

    If ``x`` is complex, return an uncertain complex number.
    
    If ``x`` is real return an uncertain real number.
        
    **Example**::
    
        >>> e = constant(math.e,label='Euler')
        >>> e
        ureal(2.718281828459045,0.0,inf, label='Euler')
    
    """
    if isinstance(x,numbers.Real):
        return UncertainReal._constant(x,label)
    elif isinstance(x,numbers.Complex):
        return UncertainComplex._constant(x,label)
    else:
        raise TypeError(
            "Cannot make a constant: {!r}".format( x )
        )
  
#----------------------------------------------------------------------------
def result(un,label=None):
    """
    Create an uncertain number to represent an intermediate result

    :arg un: :class:`~lib.UncertainReal` or :class:`~lib.UncertainComplex` or :class:`.UncertainArray`
    :arg label: str or a sequence of str
    :rtype: :class:`~lib.UncertainReal` or :class:`~lib.UncertainComplex` or :class:`.UncertainArray`
        
    When ``un`` is an array, an :class:`.UncertainArray` is returned  
    containing the intermediate uncertain number objects.

    The component of uncertainty, or the sensitivity, of an uncertain number 
    with respect to an intermediate result can be evaluated. 

    Declaring intermediate results also enables uncertain 
    numbers to be stored in an archive.
        
    .. note::
    
        This function will usually be applied to a temporary object.

    .. note::
    
        Applying this function to an elementary uncertain number does not create 
        an intermediate intermediate result. If a label is already assigned to the 
        elementary uncertain number, it will not be changed and a warning 
        will be raised, otherwise :func:`~core.result` will assign a label.
    
    **Example**::

        >>> I = ureal(1.3E-3,0.01E-3)
        >>> R = ureal(995,7)
        >>> V = result( I*R )
        >>> P = V**2/R
        >>> component(P,V)
        3.505784505642068e-05  
    
    """
    if hasattr(un,'_intermediate'):
        return un._intermediate(label)
        
    elif isinstance(un,numbers.Complex):
        # This covers any pure number type
        if label is not None:
            warnings.warn(
                "A label cannot be applied to a pure number by result():"
                " the label {!r} has been ignored".format(label),
                RuntimeWarning
            )
        
        return un
        
    else:
        raise TypeError(
            "undefined for {!r}'".format(un)
        )
          

#----------------------------------------------------------------------------
def ucomplex(z,u,df=inf,label=None,independent=True):
    """
    Create an elementary uncertain complex number

    :arg z: the value (estimate)
    :type z: complex

    :arg u: the standard uncertainty or variance
    :type u: float, 2-element or 4-element sequence

    :arg df: the degrees-of-freedom
    :type df: float

    :type label: str 
    
    :rtype: :class:`~lib.UncertainComplex`
    :raises: :exc:`ValueError` if ``df`` or ``u`` have illegal values.

    ``u`` can be a float, a 2-element or 4-element sequence.

    If ``u`` is a float, the standard uncertainty in both
    the real and imaginary components is taken to be ``u``.

    If ``u`` is a 2-element sequence, the first element is
    taken to be the standard uncertainty in the real component 
    and the second element is taken to be the standard 
    uncertainty in the imaginary component.

    If ``u`` is a 4-element sequence, the sequence is 
    interpreted as a variance-covariance matrix.

    **Examples**::

        >>> uc = ucomplex(1+2j,(.5,.5),3,label='x')
        >>> uc
        ucomplex((1+2j), u=[0.5,0.5], r=0.0, df=3.0, label=x)
   
    >>> cv = (1.2,0.7,0.7,2.2)
    >>> uc = ucomplex(0.2-.5j, cv)
    >>> variance(uc)
    VarianceCovariance(rr=1.1999999999999997, ri=0.7, ir=0.7, ii=2.2)
    
    """
    # Arguments to these math functions must be compatible with float
    # otherwise a TypeError is raised by Python
    if cmath.isnan(z) or cmath.isinf(z):
        raise ValueError("invalid: '{!r}'".format(z) )
        
    if df < 1 or math.isnan(df):
        raise ValueError("invalid dof: '{!r}'".format(df) )
        
    if is_sequence(u):
    
        case = len(u)
        
        if case == 2:
            u_r = float(u[0])
            u_i = float(u[1])
            r = None

        elif case == 4:
            u_r,cv1,cv2,u_i = u

            # nan != nan is True
            if math.isinf(cv1) or cv1 != cv2:
                raise ValueError(
                    "covariance elements not equal: {!r} and {!r}".format(cv1,cv2) 
                )
            u_r = math.sqrt(u_r)
            u_i = math.sqrt(u_i)
            r = cv1 / (u_r*u_i) if cv1 != 0 else None

            # Allow a little tolerance for numerical imprecision
            if r is not None and abs(r) > 1 + 1E-10:
                raise ValueError(
                    "invalid correlation: {!r}, cv={}".format(r,u)
                )
            if r is not None:
                # This overrides an initial assignment
                independent = False
                
        else:
            raise ValueError(
                "invalid uncertainty sequence: '{!r}'".format(u)
            )
        
    elif not math.isinf(u) and not math.isnan(u):
        u_r = u_i = float(u)
        r = None
    else:
        raise TypeError("invalid uncertainty: '{!r}'".format(u) )

    # Checking of valid uncertainty values
    # Note, comparisons with nan are always false
    if not( 0 <= u_r and u_r < inf ):
        raise ValueError("invalid real uncertainty: '{!r}'".format(u_r) )

    if not ( 0 <= u_i and u_i < inf ):
        raise ValueError("invalid imag uncertainty: '{!r}'".format(u_i) )
        
    # TODO: is this what we want? Perhaps not!
    if u_r == 0 and u_i == 0:
        return UncertainComplex._constant(complex(z),label)
    else:
        return UncertainComplex._elementary(
            complex(z),
            u_r,u_i,r,
            float(df),
            label,
            independent
        )
        
#---------------------------------------------------------------------------
# TODO: think of a better name! Perhaps `ucomplex_ensemble`
def multiple_ucomplex(x_seq,u_seq,df,label_seq=None):
    """Return a sequence of uncertain complex numbers

    :arg x_seq: a sequence of complex values
    :arg u_seq: a sequence of standard uncertainties or covariances
    :arg df: the degrees-of-freedom
    :arg label_seq: a sequence of labels for the uncertain numbers

    :rtype: a sequence of :class:`~lib.UncertainComplex`
    
    This function defines an set of uncertain complex
    numbers with the same number of degrees-of-freedom.
    
    Correlation between any pairs of these uncertain  
    numbers will not invalidate degrees-of-freedom calculations.
    (see: R Willink, *Metrologia* 44 (2007) 340-349, Sec. 4.1)
    
    **Example**::
    
        # GUM Appendix H2
        >>> values = [4.999+0j,0.019661+0j,1.04446j]
        >>> uncert = [(0.0032,0.0),(0.0000095,0.0),(0.0,0.00075)]
        >>> v,i,phi = multiple_ucomplex(values,uncert,5)
        
        >>> set_correlation(-0.36,v.real,i.real)
        >>> set_correlation(0.86,v.real,phi.imag)
        >>> set_correlation(-0.65,i.real,phi.imag)
        
        >>> z = v * exp(phi)/ i
        >>> print(z)
        (127.732(70)+219.847(296)j)
        >>> z.r
        -0.591484610818998...

    """
    if len(x_seq) != len(u_seq):
        raise RuntimeError(
            "unequal length sequences: x={!r} u={!r}".format(x_seq,u_seq)
        )
 
    if label_seq is None:
        label_seq = [None]*len(x_seq)
    elif is_sequence(label_seq):
        if len(x_seq) != len(label_seq):
            raise RuntimeError(
                "unequal length sequences: x={!r} label_seq={!r}".format(x_seq,u_seq)
            )
    else:
        raise RuntimeError(
            "invalid `label_seq`: {!r}".format(label_seq)
        )
 
    rtn = [
        # When u_i == 0 constant objects are created
        ucomplex(x_i,u_i,df,label=l_i,independent=False)
            for x_i,u_i,l_i in izip(
                x_seq,u_seq,label_seq
            )
    ]

    # Only non-constant UNs can be collected in an ensemble
    lib.complex_ensemble(
        [ 
            un_i for un_i in rtn
                if not lib._is_uncertain_complex_constant(un_i) 
        ], 
        df
    )

    # All uncertain numbers are returned, including the constants
    return rtn

#----------------------------------------------------------------------------
def set_correlation(r,arg1,arg2=None):
    """Set correlation between elementary uncertain numbers

    The input arguments can be a pair of uncertain numbers
    (the same type, real or complex), or a single
    uncertain complex number.
    
    The uncertain number arguments must be elementary 
    uncertain numbers.
    
    If the arguments have finite degrees of 
    freedom, they must be declared together using either 
    :func:`~core.multiple_ureal` or :func:`~multiple_ucomplex`.
    
    If the uncertain number arguments have infinite degrees of 
    freedom they can, alternatively, be declared by setting the 
    argument `independent=False` when calling 
    :func:`~ureal` or :func:`~ucomplex`.
        
    A :exc:`ValueError` is raised when illegal arguments are used

    When a pair of uncertain real numbers is provided,
    ``r`` is the correlation coefficient between them. 
    
    When a pair of uncertain complex number arguments is provided,
    ``r`` must be a 4-element sequence containing correlation
    coefficients between the components of the complex quantities.
    
    **Examples**::

        >>> x1 = ureal(2,1,independent=False)
        >>> x2 = ureal(5,1,independent=False)
        >>> set_correlation(.3,x1,x2)
        >>> get_correlation(x1,x2)
        0.3

        >>> z = ucomplex(1+0j,(1,1),independent=False)
        >>> z
        ucomplex((1+0j), u=[1.0,1.0], r=0.0, df=inf)
        >>> set_correlation(0.5,z)
        >>> z
        ucomplex((1+0j), u=[1.0,1.0], r=0.0, df=inf)

        >>> x1 = ucomplex(1,(1,1),independent=False)
        >>> x2 = ucomplex(1,(1,1),independent=False)
        >>> correlation_mat = (0.25,0.5,0.75,0.5)
        >>> set_correlation(correlation_mat,x1,x2)
        >>> get_correlation(x1,x2)
        CorrelationMatrix(rr=0.25, ri=0.5, ir=0.75, ii=0.5)
    
    """
    # This requires no action, because no correlation 
    # is the same as zero.
    # NB, we don't check that a correlation coefficient 
    # is being re-defined, ever! If someone were doing that 
    # and tried to set to zero, they'd be disappointed :-)
    if r == 0.0: return 
    
    # This relies on checking done in the calling functions 
    # to ensure that arguments are elementary, declared dependent,
    # and that the correlation coefficient value is OK.
    if isinstance( arg1,(UncertainReal,UncertainComplex) ):
        arg1.set_correlation(r,arg2)
    else:
        raise TypeError(
            "illegal arguments: {}, {}, {}".format( 
                repr(r), repr(arg1), repr(arg2) 
            )
        )

#---------------------------------------------------------------------------
def get_correlation(arg1,arg2=None):
    """Return correlation 
    
    The input arguments may be a pair of uncertain numbers, 
    or a single uncertain complex number.
    
    When a pair of uncertain real numbers is provided,
    the correlation between the arguments is returned as 
    a real number. 
    
    When one, or both, arguments are uncertain complex numbers,
    a :obj:`~named_tuples.CorrelationMatrix` is returned, 
    representing a 2-by-2 matrix of correlation coefficients.
    
    """
    # Return zero if numerical arguments are given
    
    # If the arg is any number type, it matches `numbers.Complex`
    if isinstance(arg1,numbers.Complex): 
        arg1 = constant(arg1)
    
    if isinstance(arg1,(UncertainReal,UncertainComplex)):
        return arg1.get_correlation(arg2)
        
    else:
        raise TypeError(
            "illegal first argument {!r}".format(arg1)
        )
        
#---------------------------------------------------------------------------
def get_covariance(arg1,arg2=None):
    """Evaluate covariance.
    
    The input arguments can be a pair of uncertain numbers, 
    or a single uncertain complex number.
    
    When a pair of uncertain real numbers is supplied,
    the correlation between the two arguments is returned 
    as a real number. 
    
    When one, or both, arguments are uncertain complex numbers,
    a :class:`~named_tuples.CovarianceMatrix` is returned, 
    representing a 2-by-2 variance-covariance matrix.
    
    """
    # If numerical arguments are given then return zero. 
    # NB if the arg is any number type, it matches `numbers.Complex`
    if isinstance(arg1,numbers.Complex): 
        arg1 = constant(arg1)
    
    if isinstance( arg1,(UncertainReal,UncertainComplex) ):
        return arg1.get_covariance(arg2)
        
    else:
        raise TypeError(
            "illegal first argument {!r}".format(arg1)
        )
        
#---------------------------------------------------------------------------
def log(x):
    """
    Uncertain number natural logarithm

    .. note::
        In the complex case there is one branch cut,
        from 0 along the negative real axis to :math:`-\infty`,
        continuous from above.
        
    """
    try:
        return x._log()
    except AttributeError:
        if isinstance(x,numbers.Real):
            return math.log(x)
        elif isinstance(x,numbers.Complex):
            return cmath.log(x)
        else:
            raise TypeError(
                "illegal argument: {!r}".format(x)
            )
    
#---------------------------------------------------------------------------
def log10(x):
    """
    Uncertain number common logarithm (base-10)

    .. note::
        In the complex case there is one branch cut,
        from 0 along the negative real
        axis to :math:`-\infty`, continuous from above.
        
    """
    try:
        return x._log10()
    except AttributeError:
        if isinstance(x,numbers.Real):
            return math.log10(x)
        elif isinstance(x,numbers.Complex):
            return cmath.log10(x)
        else:
            raise TypeError(
                "illegal argument: {!r}".format(x)
            )
    
#---------------------------------------------------------------------------
def exp(x):
    """
    Uncertain number exponential function

    """
    try:
        return x._exp()
    except AttributeError:
        if isinstance(x,numbers.Real):
            return math.exp(x)
        elif isinstance(x,numbers.Complex):
            return cmath.exp(x)
        else:
            raise TypeError(
                "illegal argument: {!r}".format(x)
            )
    
#---------------------------------------------------------------------------
def pow(x,y):
    """
    Uncertain number power function
    
    Raises ``x`` to the power of ``y``
    
    """
    return x**y
        
#---------------------------------------------------------------------------
def sqrt(x):
    """
    Uncertain number square root function

    .. note::
        In the complex case there is one branch cut,
        from 0 along the negative real
        axis to :math:`-\infty`, continuous from above.
        
    """
    try:
        return x._sqrt()
    except AttributeError:
        if isinstance(x,numbers.Real):
            return math.sqrt(x)
        elif isinstance(x,numbers.Complex):
            return cmath.sqrt(x)
        else:
            raise TypeError(
                "illegal argument: {!r}".format(x)
            )

#----------------------------------------------------------------------------
def sin(x):
    """
    Uncertain number sine function

    """
    try:
        return x._sin()
    except AttributeError:
        if isinstance(x,numbers.Real):
            return math.sin(x)
        elif isinstance(x,numbers.Complex):
            return cmath.sin(x)
        else:
            raise TypeError(
                "illegal argument: {!r}".format(x)
            )

#----------------------------------------------------------------------------
def cos(x):
    """
    Uncertain number cosine function

    """
    try:
        return x._cos()
    except AttributeError:
        if isinstance(x,numbers.Real):
            return math.cos(x)
        elif isinstance(x,numbers.Complex):
            return cmath.cos(x)
        else:
            raise TypeError(
                "illegal argument: {!r}".format(x)
            )
#----------------------------------------------------------------------------
def tan(x):
    """
    Uncertain number tangent function

    """
    try:
        return x._tan()
    except AttributeError:
        if isinstance(x,numbers.Real):
            return math.tan(x)
        elif isinstance(x,numbers.Complex):
            return cmath.tan(x)
        else:
            raise TypeError(
                "illegal argument: {!r}".format(x)
            )

#---------------------------------------------------------------------------
def asin(x):
    """
    Uncertain number arcsine function

    .. note::
        In the complex case there are two branch cuts: one extends
        right, from 1 along the real axis to :math:`\infty`,
        continuous from below; the other extends left, from -1 along
        the real axis to :math:`-\infty`, continuous from above.
        
    """
    try:
        return x._asin()
    except AttributeError:
        if isinstance(x,numbers.Real):
            return math.asin(x)
        elif isinstance(x,numbers.Complex):
            return cmath.asin(x)
        else:
            raise TypeError(
                "illegal argument: {!r}".format(x)
            )
#---------------------------------------------------------------------------
def acos(x):
    """
    Uncertain number arc-cosine function

    .. note::
        In the complex case there are two branch cuts: one extends
        right, from 1 along the real axis to :math:`\infty`, continuous
        from below; the other extends left, from -1 along the real axis
        to :math:`-\infty`, continuous from above.
        
    """
    try:
        return x._acos()
    except AttributeError:
        if isinstance(x,numbers.Real):
            return math.acos(x)
        elif isinstance(x,numbers.Complex):
            return cmath.acos(x)
        else:
            raise TypeError(
                "illegal argument: {!r}".format(x)
            )
            
#---------------------------------------------------------------------------
def atan(x):
    """
    Uncertain number arctangent function

    .. note::
    
        In the complex case there are two branch cuts:
        One extends from :math:`\mathrm{j}` along the imaginary axis to
        :math:`\mathrm{j}\infty`, continuous from the right.
        The other extends from :math:`-\mathrm{j}` along the imaginary
        axis to :math:`-\mathrm{j}\infty`, continuous from the left.
        
    """
    try:
        return x._atan()
    except AttributeError:
        if isinstance(x,numbers.Real):
            return math.atan(x)
        elif isinstance(x,numbers.Complex):
            return cmath.atan(x)
        else:
            raise TypeError(
                "illegal argument: {!r}".format(x)
            )
            
#----------------------------------------------------------------------------
def atan2(y,x):
    """
    Two-argument uncertain number arctangent function    

    :arg x: abscissa
    :type x: :class:`~lib.UncertainReal`
    :arg y: ordinate
    :type y: :class:`~lib.UncertainReal`
    
    .. note::   this function is not defined for uncertain complex numbers
                (use :func:`phase`)

    **Example**::

        >>> x = ureal(math.sqrt(3)/2,1)
        >>> y = ureal(0.5,1)
        >>> theta = atan2(y,x)
        >>> theta
        ureal(0.5235987755982989,1.0,inf)
        >>> math.degrees( theta.x )
        30.000000000000004
    
    """
    try:
        return y._atan2(x)
    except AttributeError:
        try:
            return x._ratan2(y)
        except AttributeError:
            if isinstance(x,numbers.Real) and isinstance(y,numbers.Real):
                return math.atan2(y,x)
            else:
                raise TypeError(
                    "illegal arguments: x={!r} y={!r}".format(x,y)
                )

#---------------------------------------------------------------------------
def sinh(x):
    """
    Uncertain number hyperbolic sine function

    """
    try:
        return x._sinh()
    except AttributeError:
        if isinstance(x,numbers.Real):
            return math.sinh(x)
        elif isinstance(x,numbers.Complex):
            return cmath.sinh(x)
        else:
            raise TypeError(
                "illegal argument: {!r}".format(x)
            )
#---------------------------------------------------------------------------
def cosh(x):
    """
    Uncertain number hyperbolic cosine function

    """
    try:
        return x._cosh()
    except AttributeError:
        if isinstance(x,numbers.Real):
            return math.cosh(x)
        elif isinstance(x,numbers.Complex):
            return cmath.cosh(x)
        else:
            raise TypeError(
                "illegal argument: {!r}".format(x)
            )
    
#---------------------------------------------------------------------------
def tanh(x):
    """
    Uncertain number hyperbolic tangent function

    """
    try:
        return x._tanh()
    except AttributeError:
        if isinstance(x,numbers.Real):
            return math.tanh(x)
        elif isinstance(x,numbers.Complex):
            return cmath.tanh(x)
        else:
            raise TypeError(
                "illegal argument: {!r}".format(x)
            )
            
#---------------------------------------------------------------------------
def asinh(x):
    """
    Uncertain number hyperbolic arcsine function

    .. note::
    
        In the complex case there are two branch cuts: one extends
        from :math:`\mathrm{j}` along the imaginary axis to
        :math:`\mathrm{j}\infty`, continuous from the right;
        the other extends from :math:`-\mathrm{j}` along the
        imaginary axis to :math:`-\mathrm{j}\infty`, continuous
        from the left.
        
    """
    try:
        return x._asinh()
    except AttributeError:
        if isinstance(x,numbers.Real):
            return math.asinh(x)
        elif isinstance(x,numbers.Complex):
            return cmath.asinh(x)
        else:
            raise TypeError(
                "illegal argument: {!r}".format(x)
            )

#---------------------------------------------------------------------------
def acosh(x):
    """
    Uncertain number hyperbolic arc-cosine function

    .. note::
        In the complex case there is one branch cut,
        extending left from 1 along the
        real axis to :math:`-\infty`, continuous from above.
        
    """
    try:
        return x._acosh()
    except AttributeError:
        if isinstance(x,numbers.Real):
            return math.acosh(x)
        elif isinstance(x,numbers.Complex):
            return cmath.acosh(x)
        else:
            raise TypeError(
                "illegal argument: {!r}".format(x)
            )

#---------------------------------------------------------------------------
def atanh(x):
    """
    Uncertain number hyperbolic arctangent function

    .. note::
        In the complex case there are two branch cuts:
        one extends from 1 along the real axis to :math:`\infty`,
        continuous from below; the other extends from -1
        along the real axis to :math:`-\infty`, continuous
        from above.
        
    """
    try:
        return x._atanh()
    except AttributeError:
        if isinstance(x,numbers.Real):
            return math.atanh(x)
        elif isinstance(x,numbers.Complex):
            return cmath.atanh(x)
        else:
            raise TypeError(
                "illegal argument: {!r}".format(x)
            )
            
#----------------------------------------------------------------------------
def magnitude(x):
    """
    Return the magnitude of ``x``

    .. note::

        If ``x`` is not an uncertain number type,
        returns :func:`abs(x)<abs>`.    
    
    """
    try:
        return x._magnitude()
    except AttributeError:
        return abs(x)

#---------------------------------------------------------------------------
def phase(z):
    """
    :arg z: an uncertain complex number    
    :type z: :class:`~lib.UncertainComplex`

    :returns:   the phase in radians
    :rtype:     :class:`~lib.UncertainReal`
    
    """
    try:
        return z._phase()
    except AttributeError:
        if isinstance(z,numbers.Complex):
            return cmath.phase(z)
        else:
            raise TypeError(
                "illegal argument: {!r}".format(z)
            )

#---------------------------------------------------------------------------
def mag_squared(x):
    """
    Return the squared magnitude of ``x``.

    .. note::
    
        If ``x`` is an uncertain number, the magnitude
        squared is returned as an uncertain real number, 
        otherwise :func:``abs(x)**2`` is returned.
    
    """
    try:
        return x._mag_squared()
    except AttributeError:
        return abs(x)**2

# import here to avoid circular imports
from GTC import linear_algebra
la = linear_algebra

#============================================================================    
if __name__ == "__main__":
    import doctest       
    # from GTC import *
    doctest.testmod(  optionflags=doctest.NORMALIZE_WHITESPACE )



"""
Copyright (c) 2018, Measurement Standards Laboratory of New Zealand.

"""
import uuid
import weakref
try:
    long  # Python 2
except NameError:
    long = int

from GTC.nodes import (
    Leaf,
    Node
)

__all__ = (
    'Context',
    '_context'
)

#----------------------------------------------------------------------------
class Context(object):

    """
    A ``Context`` object the creation of uncertain numbers 
    """
    
    def __init__(self,id=None):
        
        # A UUID4 identifier is the default.
        # However, the user can enter a specific ID if 
        # required (there is no guarantee then that 
        # Context IDs will be unique - user beware!)
        self._id = uuid.uuid4().int if id is None else id
    
        self._elementary_id_counter = long(0)
        self._intermediate_id_counter = long(0)
  
        # Caching to avoid duplication when unpacking archives 
        self._registered_leaf_nodes = weakref.WeakValueDictionary()
        self._registered_intermediate_nodes = weakref.WeakValueDictionary()
          
    #------------------------------------------------------------------------
    def _next_elementary_id(self):
        """
        Return an ID number

        Each elementary uncertain number has a 
        unique ID. This is composed of the 
        context's ID and an integer.

        Returns
        -------
        ID : tuple(context_ID,integer)
        
        """
        self._elementary_id_counter += 1
        return (
            self._id,
            self._elementary_id_counter
        )

    #------------------------------------------------------------------------
    def _next_intermediate_id(self):
        """
        Return an ID number

        Each intermediate uncertain number has
        a unique ID composed of the context's ID
        and an integer.

        Returns
        -------
        ID : tuple(context_ID,integer)
        
        
        ..Note:
        
            The ID generated here has the same  
            format as elementary UIDs, but they  
            are independent: values can overlap!
            
        """
        self._intermediate_id_counter += 1
        return (
            self._id,
            self._intermediate_id_counter
        )
 
    #------------------------------------------------------------------------
    def new_leaf(self,uid,label,u,df,independent):
        """
        Return a new ``Leaf`` node unless one with the same uid exists
        
        """        
        if uid in self._registered_leaf_nodes:
            # If the node found is indistinguishable from the new node 
            # then quietly ignore the request 
            l = self._registered_leaf_nodes[uid]
            OK = (
                label == l.label and 
                u == l.u and 
                df == l.df and 
                independent == l.independent
            )
            if not OK:
                raise RuntimeError(
                    "the Leaf node uid({}) is in use already".format(uid)
                )
        else:          
            l = Leaf(uid,label,u,df,independent)
            self._registered_leaf_nodes[uid] = l
            
        return l 
            
    #------------------------------------------------------------------------
    def new_node(self,uid,label,u,df):
        """
        Return a new ``Node`` unless one with the same uid exists
        
        """
        
        # Prior to v1.3.5, degrees of freedom was not buffered in a Node.
        # Now, to maintain backward comparability, a value ``None`` 
        # will be inserted for backward compatibility when needed.
        # See the ``shim_1_3_3`` function in archive.py.        

        if uid in self._registered_intermediate_nodes:
            raise RuntimeError(
                "intermediate node uid({}), '{}', u={}, df={} is used".format(
                    uid,label,u,df
                )
            )
        else:          
            n = Node(uid,label,u,df)
            self._registered_intermediate_nodes[uid] = n

        return n


#----------------------------------------------------------------------------
# A default context
#
_context = Context()



"""
Reporting functions
-------------------

    *   The function :func:`budget` produces an uncertainty budget.
    *   The function :func:`k_factor` returns the coverage factor 
        used for real-valued problems (based on the Student-t distribution). 
    *   The function :func:`k_to_dof` returns the degrees of freedom 
        corresponding to a given coverage factor and coverage probability.
    *   The function :func:`k2_factor_sq` returns   
        coverage factor squared for the complex-valued problem. 
    *   The function :func:`k2_to_dof` returns the degrees of freedom 
        corresponding to a given coverage factor and coverage probability
        in complex-valued problems.
    *   Functions :func:`u_bar` and :func:`v_bar` return summary values 
        for matrix results associated with 2-D uncertainty.

Uncertainty functions
---------------------

    *   The function :func:`u_component` returns the signed 
        component of uncertainty in one uncertain number 
        due to uncertainty in another.
        
    *   The function :func:`sensitivity` returns the partial 
        derivative of one uncertain number with respect to another.
        This is often called a sensitivity coefficient.

Type functions
--------------

    *   The function :func:`is_ureal` can be used to
        identify uncertain real numbers. 
    *   The function :func:`is_ucomplex` can be used to
        identify uncertain complex numbers. 

Module contents
---------------

"""
from __future__ import division     # True division

import math
import numbers

from operator import attrgetter as getter
from functools import reduce

from scipy import special, optimize

try:
    from itertools import izip  # Python 2
except ImportError:
    izip = zip
    long = int

from GTC.lib import (
    UncertainReal,
    UncertainComplex,
    _is_uncertain_real_constant,
    _is_uncertain_complex_constant
)
from GTC.named_tuples import (
    ComponentOfUncertainty, 
    Influence
)
from GTC.vector import extend_vector

from GTC import (
    is_sequence,
    inf,
    inf_dof,
)

__all__ = (
    'budget',
    'k_factor',
    'k_to_dof',
    'k2_factor_sq',
    'k2_to_dof',
    'u_component',
    'sensitivity',
    'is_ureal',
    'is_ucomplex',
    'v_bar',
    'u_bar',
)

#--------------------------------------------------------------------------
uid_str = lambda id: "{0[0]:d}_{0[1]:d}".format(id)

#--------------------------------------------------------------------------
def is_ureal(x):
    """Return ``True`` if ``x`` is an uncertain real number
    
    **Example**::

        >>> x = ureal(1,1)
        >>> reporting.is_ureal(x)
        True
        
    """
    return isinstance(x,UncertainReal)
    
#--------------------------------------------------------------------------
def is_ucomplex(z):
    """Return ``True`` if ``z`` is an uncertain complex number
    
    **Example**::

        >>> z = ucomplex(1+2j,(0.1,0.2))
        >>> reporting.is_ucomplex(z)
        True

    """
    return isinstance(z,UncertainComplex)

#------------------------------------------------------------
def _df_k2(k2,p,nu1,TOL):
    """
    Return `nu2` such that the integral of 
    F(nu1,nu2) from -infty to `x` is `p`
    
    `x` is k2**2 * nu2/ ( nu1*(nu2+1) )
    
    """
    # We have `k2` the integral limit, so `pf` gives us `p`
    # we must vary the `nu2` argument until the
    # returned value equals `p`.
    # `fdtr` returns the integral of F probability density from -infty to `x`
    def fn(nu2):
        x = k2**2 * nu2/ ( nu1*(nu2+1) )  
        # return pf(x,nu1,nu2) - p 
        return special.fdtr(nu1,nu2,x) - p 
    
    # dof here is nu2-1 and cannot be less than 2
    # This setting of `lo` is not a strict bound, because
    # the calculation will succeed, we just don't want to 
    # go there.
    
    lo = 1 - 1E-3   
    fn_lo = fn(lo)
    if fn_lo > 0.0:
        raise RuntimeError(
            "dof < 2 cannot be calculated"
        )
        
    upper_limit = (20,50,1E2,1E3,inf_dof)
    for hi in upper_limit:
        if fn(hi) > 0.0: 
            return optimize.ridder(fn,lo,hi)
        else:    
            lo = hi
        
    return inf       
#----------------------------------------------------------------------------
def k2_to_dof(k2,p=95):
    """Return the dof corresponding to a bivariate coverage factor `k2`  
    
    :arg k2: coverage factor (>0)
    :arg p: coverage probability (%)
    :type k2: float
    :type p: int or float

    Evaluates a number of degrees-of-freedom given a coverage 
    factor for an elliptical uncertainty region with coverage 
    probability ``p`` based on the F-distribution.
    
    **Example**::

        >>> reporting.k2_to_dof(2.6,95)
        34.35788424389927
        
    """
    if k2 <= 0:
        raise RuntimeError( "invalid k:  {}".format(k2) ) 
    if p <= 0 or p >= 100:
        raise RuntimeError( "invalid p: {}".format(p) )
    else:
        p = p / 100.0     

    return _df_k2(k2,p,2,1E-7) + 1

#----------------------------------------------------------------------------
def k2_factor_sq(df=inf,p=95):
    """Return a squared coverage factor for an elliptical uncertainty region

    :arg df: the degrees-of-freedom (>=2)
    :arg p: the coverage probability (%)
    :type df: float
    :type p: int or float

    Evaluates the square of the coverage factor for an elliptical uncertainty 
    region with coverage probability ``p``  and ``df`` degrees of freedom
    based on the F-distribution.
    
    **Example**::

        >>> reporting.k2_factor_sq(3)
            56.99999999999994
    
    """
    p = p / 100.0
    
    if df > inf_dof:
        return -2.0 * math.log(1-p)
        
    elif(df>1):   
        # norm = l * (n-1) / (n - l) in the general
        # 'l'-dimensional case for 'n' observations
        # here l = 2, df = n-1
        norm = 2*df / (df-1)
        
        # `fdtri` is the inverse of the cumulative F distribution
        # returning `x` such that `fdtr(dfn, dfd, x) = p`
        return norm*special.fdtri(2.0,df-1.0,p)
        
    else:
        raise RuntimeError("invalid df={!r}".format( df ) )
 
#----------------------------------------------------------------------------
def k_factor(df=inf,p=95):
    """Return the a coverage factor for an uncertainty interval

    :arg df: the degrees-of-freedom (>1)
    :arg p: the coverage probability (%)
    :type df: float
    :type p: int or float

    Evaluates the coverage factor for an uncertainty interval
    with coverage probability ``p`` and degrees-of-freedom ``df``
    based on the Student t-distribution. 
    
    **Example**::
    
        >>> reporting.k_factor(3)
        3.182446305284263

    """
    if p <= 0.0 or p >= 100.0:
        raise RuntimeError( "invalid p: {}".format( p ) )
    
    p = (1.0 + p/100.0)/2.0
    
    if df > inf_dof:
        # inverse cumulative Gaussian distribution
        return special.ndtri(p)
    elif df >= 1:
        # inverse cumulative Student-t distribution
        return special.stdtrit(df,p)
    else:
        raise RuntimeError( "invalid df: {}".format( df ) )
   
#----------------------------------------------------------------------------
def k_to_dof(k,p=95):
    """Return the dof corresponding to a univariate coverage factor `k` 
    
    :arg k: coverage factor (>0)
    :arg p: coverage probability (%)
    :type k: float
    :type p: int or float

    Evaluates the degrees-of-freedom given a coverage factor for 
    an uncertainty interval with coverage probability ``p``
    based on the Student t-distribution.
    
    **Example**::

        >>> reporting.k_to_dof(2.0,95)
        60.43756442698591
        
    """
    if k <= 0:
        raise RuntimeError( "invalid k:  {}".format( k ) )  
    if p <= 0 or p >= 100:
        raise RuntimeError( "invalid p: {}".format( p ) )
    else:
        p = (1.0 + p/100.0)/2.0         
        df = special.stdtridf(p,k) 
        
        return df if df < inf_dof else inf 

#----------------------------------------------------------------------------
def sensitivity(y,x):
    """Return the first partial derivative of ``y`` with respect to ``x``

    :arg y: :class:`~lib.UncertainReal` or :class:`~lib.UncertainComplex` or :class:`.UncertainArray` 
    :arg x: :class:`~lib.UncertainReal` or :class:`~lib.UncertainComplex` or :class:`.UncertainArray` 
    
    If ``x`` and ``y`` are uncertain real numbers, return a float. 

    If ``y`` or ``x`` is an uncertain complex number, return 
    a 4-element sequence of float, representing the Jacobian matrix.

    When ``x`` and ``y`` are arrays, an :class:`.UncertainArray` 
    is returned containing the results of applying this function 
    to the array elements.

    Otherwise, return 0.
    
    .. versionadded:: 1.1

    **Example**::

        >>> x = ureal(3,1)
        >>> y = 3 * x
        >>> reporting.sensitivity(y,x)
        3.0

        >>> q = ucomplex(2,1)
        >>> z = magnitude(q)    # uncertain real numbers
        >>> reporting.sensitivity(z,q)
        JacobianMatrix(rr=1.0, ri=0.0, ir=0.0, ii=0.0)
        
        >>> r = ucomplex(3,1)
        >>> z = q * r
        >>> reporting.sensitivity(z,q)
        JacobianMatrix(rr=3.0, ri=-0.0, ir=0.0, ii=3.0)
        
    """
    # There are three types that define a `sensitivity` method: 
    # ~uncertain_array.UncertainArray, UncertainReal and UncertainComplex. 
    # These are all potential types for `y`. 
    # The types for `x` include these three as well as numbers.
    # 
    if hasattr(y,'sensitivity'):
        return y.sensitivity(x)
    else:
        raise RuntimeError(
            "An uncertain number is expected: {!r}".format(y)
        ) 
        
#----------------------------------------------------------------------------
def u_component(y,x):
    """Return the component of uncertainty in ``y`` due to ``x``
    
    :arg y: :class:`~lib.UncertainReal` or :class:`~lib.UncertainComplex` or :class:`.UncertainArray` 
    :arg x: :class:`~lib.UncertainReal` or :class:`~lib.UncertainComplex` or :class:`.UncertainArray` 
    
    If ``x`` and ``y`` are uncertain real numbers, return a float. 

    If ``y`` or ``x`` is an uncertain complex number, return 
    a 4-element sequence of float, containing the
    components of uncertainty.

    When ``x`` and ``y`` are arrays, an :class:`uncertain_array.UncertainArray` 
    is returned containing the results of applying this function 
    to the array elements.

    Otherwise, return 0.

    **Example**::

        >>> x = ureal(3,1)
        >>> y = 3 * x
        >>> reporting.u_component(y,x)
        3.0

        >>> q = ucomplex(2,1)
        >>> z = magnitude(q)    # uncertain real numbers
        >>> reporting.u_component(z,q)
        ComponentOfUncertainty(rr=1.0, ri=0.0, ir=0.0, ii=0.0)
        
        >>> r = ucomplex(3,1)
        >>> z = q * r
        >>> reporting.u_component(z,q)
        ComponentOfUncertainty(rr=3.0, ri=-0.0, ir=0.0, ii=3.0)
        
    """
    if hasattr(y,'u_component'):
        return y.u_component(x)
    else:
        raise RuntimeError(
            "An uncertain number is expected: {!r}".format(y)
        )

#----------------------------------------------------------------------------
def u_bar(ucpt):
    """Return the magnitude of a component of uncertainty
    
    :arg ucpt: a component of uncertainty
    :type ucpt: float or 4-element sequence of float

    If ``ucpt`` is a sequence, return the root-sum-square 
    of the elements divided by :math:`\sqrt{2}`

    If ``ucpt`` is a number, return the absolute value.

    **Example**::

        >>> x1 = 1-.5j
        >>> x2 = .2+7.1j
        >>> z1 = ucomplex(x1,1)
        >>> z2 = ucomplex(x2,1)
        >>> y = z1 * z2
        >>> dy_dz1 = reporting.u_component(y,z1)
        >>> dy_dz1
        ComponentOfUncertainty(rr=0.2, ri=-7.1, ir=7.1, ii=0.2)
        >>> reporting.u_bar(dy_dz1)
        7.102816342831905
    
    """
    if is_sequence(ucpt):
        if len(ucpt) != 4:
            raise RuntimeError(
                "need a 4-element sequence, got: {!r}".format(ucpt)
            )
           
        return math.sqrt( reduce(lambda y,x: y + x*x,ucpt,0) / 2 )
        
    elif isinstance(ucpt,numbers.Real):
        return abs(ucpt)
        
    else:
        raise RuntimeError(
            "need a 4-element sequence or float, got: {!r}".format(ucpt)
        )
        
#----------------------------------------------------------------------------
def v_bar(cv):
    """Return the trace of ``cv`` divided by 2 
    
    :arg cv: a variance-covariance matrix
    :type cv: 4-element sequence of float
    
    :returns: float

    **Example**::
    
        >>> x1 = 1-.5j
        >>> x2 = .2+7.1j
        >>> z1 = ucomplex(x1,(1,.2))
        >>> z2 = ucomplex(x2,(.2,1))
        >>> y = z1 * z2
        >>> y.v
        VarianceCovariance(rr=2.3464, ri=1.8432, ir=1.8432, ii=51.4216)
        >>> reporting.v_bar(y.v)
        26.884

    """
    assert len(cv) == 4,\
           "'%s' a 4-element sequence is needed" % type(cv)

    return (cv[0] + cv[3]) / 2.0

#----------------------------------------------------------------------------
def budget(y,**kwargs):
    """Return a sequence of label-component of uncertainty pairs

    arg:
        y (:class:`~lib.UncertainReal` or :class:`~lib.UncertainComplex`):  an uncertain number

    keyword args:
        | influences: a sequence of uncertain numbers
        | key (str): sorting key ('u' or 'label')
        | reverse (bool): sorting order (forward or reverse)
        | trim (float): control smallest reported magnitudes 
        | max_number (int): return no more than ``max_number`` components
        | intermediate (bool): report intermediate components
    
    A sequence of :obj:`~named_tuples.Influence` namedtuples is 
    returned, each with the attributes ``label`` and ``u`` for a 
    component of uncertainty (see :func:`~core.component`). 

    The keyword argument ``influences`` can select specific influences
    to be reported.

    The keyword argument ``key`` sets the order of the sequence
    by the component of uncertainty or the label (``u`` or ``label``).

    The keyword argument ``reverse`` controls the sense of ordering.
    
    The keyword argument ``trim`` can be used to set a minimum relative 
    magnitude of components returned. The components of uncertainty greater 
    than ``trim`` times the largest component will be reported. 
    Set ``trim=0`` for a complete list.

    The keyword argument ``max_number`` can be used to restrict the 
    number of components returned.  
    
    The keyword argument ``intermediate`` allows all the components 
    of uncertainty with respect to all intermediate results to be reported.

    **Examples**::

        >>> x1 = ureal(1,1,label='x1')
        >>> x2 = ureal(2,0.5,label='x2')
        >>> x3 = ureal(3,0.1,label='x3')
        >>> y = (x1 - x2) / x3
        >>> for l,u in reporting.budget(y):
        ... 	print("{0}: {1:G}".format(l,u))
        ... 	
        x1: 0.333333
        x2: 0.166667
        x3: 0.0111111
        
        >>> for l,u in reporting.budget(y,reverse=False):
        ... 	print("{0}: {1:G}".format(l,u))
        ... 	
        x3: 0.0111111
        x2: 0.166667
        x1: 0.333333
 
        >>> y1 = result(x1 + x2,label='y1')
        >>> y2 = result(x2 + x3,label='y2')
        >>> for l,u in reporting.budget(y1 + y2,intermediate=True):
        ... 	print("{0}: {1:G}".format(l,u))
        ... 
        y1: 1.11803
        y2: 0.509902
      
    .. versionchanged::
        Added the `intermediate` keyword argument. 
        
    """  
    # Keyword options
    influences = kwargs.get('influences') 
    key = kwargs.get('key', 'u') 
    reverse = kwargs.get('reverse', True)
    trim = kwargs.get('trim', 0.01)
    max_number = kwargs.get('max_number')
    intermediate = kwargs.get('intermediate', False)
    
    # Some combinations are incompatible
    if intermediate and influences is not None:
        raise RuntimeError(
            "'influences' cannot be specified when 'intermediate' is True"
        )
    
    if isinstance(y,UncertainReal):
        if influences is None and not intermediate:
            nodes = y._u_components.keys()
            labels = [ n_i.label 
                        if n_i.label is not None else "{}".format(n_i.uid) 
                           for n_i in nodes ]
            values = [ math.fabs( u ) for u in y._u_components.itervalues() ]
            
            nodes = y._d_components.keys()
            labels += [ n_i.label 
                        if n_i.label is not None else "{}".format(n_i.uid) 
                           for n_i in nodes ]
            values += [ math.fabs( u ) for u in y._d_components.itervalues() ]
            
        elif intermediate:
            # The argument 'y' could be in the list 
            n_y_uid = y._node.uid if y.is_intermediate else 0
            
            labels = []
            values = []
            for n_i,u_i in y._i_components.iteritems():
                if n_i.uid == n_y_uid: continue    # Do not include 'y' itself
                    
                labels.append( 
                    n_i.label if n_i.label is not None else "{}".format(n_i.uid) 
                )
                values.append( math.fabs( u_i ) )
            
        elif influences is not None:
            labels = []
            values = []
            for i in influences:
                if isinstance(i,UncertainReal):
                    labels.append( i.label )
                    values.append( math.fabs(u_component(y,i)) ) 
                    
                elif isinstance(i,UncertainComplex):
                    labels.append( i.real.label )
                    values.append( math.fabs(u_component(y,i.real)) ) 
                    labels.append( i.imag.label )
                    values.append( math.fabs(u_component(y,i.imag)) ) 
                else:
                    raise RuntimeError(
                        "unexpected type: '{!r}'".format( i )
                    )
        else:
            assert False,"should never occur"
            
        if len(values):
            cut_off = max(values) * float(trim)
            this_budget = [ Influence( label=n, u=u )
                            for (u,n) in izip(values,labels) if u >= cut_off ]
        else:
            this_budget = [ ]
        
    elif isinstance(y,UncertainComplex):        
        if influences is None and not intermediate:
            
            # Ensure that the influence vectors have the same keys
            re = extend_vector(y.real._u_components, y.real._d_components)
            re = extend_vector(re,y.imag._u_components)
            re = extend_vector(re,y.imag._d_components)
    
            im = extend_vector(y.imag._u_components, y.imag._d_components)
            im = extend_vector(im,y.real._u_components)
            im = extend_vector(im,y.real._d_components)

            try:
                labels = []
                values = []
                it_re = re.iteritems()
                it_im = im.iteritems()
                
                while True:
                    ir_0,ur_0 = next(it_re)
                    ii_0,ui_0 = next(it_im)

                    if hasattr(ir_0,'complex'):
                        
                        # The next item is always the imaginary component
                        ir_1,ur_1 = next(it_re)
                        ii_1,ui_1 = next(it_im)
                        
                        # Reduce the 4 components of uncertainty 
                        # to a summary value
                        u=u_bar([ur_0,ur_1,ui_0,ui_1])
                        
                        if ir_0.label is None:
                            # No label assigned, report uids
                            label = "uid({},{})".format(
                                uid_str(ir_0.uid),uid_str(ii_0.uid)
                            )
                        else:
                            # take the trailing _re off the real label
                            # to label the complex influence
                            label = ir_0.label[:-3]
                            
                            labels.append(label)
                            values.append(u)

                    else:
                        # Report the component wrt a real influence
                        # this is still a matrix, which is then reduced 
                        # to a summary value
                        u=u_bar([ur_0,0,ui_0,0])
                        
                        if ir_0.label is None:
                            label = "uid({})".format( uid_str(ir_0.uid) )
                        else:
                            label = ir_0.label 
                            
                        labels.append(label)
                        values.append(u)
                        
            except StopIteration:
                pass
                
        elif intermediate:
            # The argument 'y' could be in the list 
            if y.is_intermediate:
                n_yr_uid = y.real._node.uid
                n_yi_uid = y.imag._node.uid
            else:
                n_yr_uid = 0
                n_yi_uid = 0
                
            # Ensure that the influence vectors have the same keys
            re = extend_vector(y.real._i_components,y.imag._i_components)    
            im = extend_vector(y.imag._i_components,y.real._i_components)

            try:
                labels = []
                values = []
                it_re = re.iteritems()
                it_im = im.iteritems()
                
                while True:
                    ir_0,ur_0 = next(it_re)
                    ii_0,ui_0 = next(it_im)

                    if hasattr(ir_0,'complex'):
                        
                        # The next item is always the imaginary component
                        ir_1,ur_1 = next(it_re)
                        ii_1,ui_1 = next(it_im)
                        
                        # Skip these real and imaginary components of 'y'
                        if ir_0.uid == n_yr_uid or ii_0.uid == n_yi_uid:
                            continue                        
                        
                        # Reduce the 4 components of uncertainty 
                        # to a summary value
                        u=u_bar([ur_0,ur_1,ui_0,ui_1])
                        
                        if ir_0.label is None:
                            # No label assigned, report uids
                            label = "uid({},{})".format(
                                uid_str(ir_0.uid),uid_str(ii_0.uid)
                            )
                        else:
                            # take the trailing _re off the real label
                            # to label the complex influence
                            label = ir_0.label[:-3]
                            
                            labels.append(label)
                            values.append(u)

                    else:
                        # Report the component wrt a real influence
                        # this is still a matrix, which is then reduced 
                        # to a summary value
                        u=u_bar([ur_0,0,ui_0,0])
                        
                        if ir_0.label is None:
                            label = "uid({})".format( uid_str(ir_0.uid) )
                        else:
                            label = ir_0.label 
                            
                        labels.append(label)
                        values.append(u)
                        
            except StopIteration:
                pass
                
        elif influences is not None:
            labels = [ i.label for i in influences ]
            values = [ u_bar( u_component(y,i) ) for i in influences ]

        else:
            assert False,"should never occur"

        if len(values):
            cut_off = max(values) * float(trim)
            this_budget = [ 
                Influence( label=n, u=u ) 
                    for (u,n) in izip(values,labels) 
                        if u >= cut_off  
            ]
            
        else:   
            this_budget = []
    else:
        this_budget = []

    if key is not None:
        this_budget.sort( key=getter(key),reverse=reverse )
    
    if max_number is not None and len(this_budget) > max_number:
        this_budget = this_budget[:max_number]
        
    return this_budget
        
#============================================================================
if __name__ == "__main__":
    import doctest
    from GTC import *
    doctest.testmod(  optionflags=doctest.NORMALIZE_WHITESPACE )


import unittest
import os

from GTC import *

#-----------------------------------------------------
class TestArchiveJSONFilev133(unittest.TestCase):

    """
    Use a reference file created using GTC v 1.3.3 
    to make sure we maintain backward compatibility.
    """
    
    def test(self):
        
        fname = 'ref_file_v_1_3_3.json'
        
        wdir =  os.path.dirname(__file__)
        path = os.path.join(wdir,fname)

        with open(path,'r') as f:
            ar = persistence.load_json(f)
            f.close()

        x1, y1, z1 = ar.extract('x','y','z')
        
        self.assertEqual( repr(x1), repr(ureal(1,1)) )
        self.assertEqual( repr(y1), repr(ureal(2,1)) )
        self.assertEqual( repr(z1), repr( ureal(1,1) + ureal(2,1) ) )

        z1 = ar.extract('z1')
        
        self.assertEqual( 
            repr(z1), 
            repr( 
                ureal(1,1,3) + ureal(2,1,4) 
            ) 
        )

        x1, y1, z1 = ar.extract('x2','y2','z2')

        x = ureal(1,1,independent=False)
        y = ureal(2,1,independent=False)
        r = 0.5
        set_correlation(r,x,y)

        self.assertEqual( repr(x1), repr(x) )
        self.assertEqual( repr(y1), repr(y) )
        self.assertEqual( repr(z1), repr(x+y) )

        x1, y1, z1 = ar.extract('x3','y3','z3')

        x,y = multiple_ureal([1,2],[1,1],4)     
        r = 0.5
        set_correlation(r,x,y)
        z = result( x + y )            
        self.assertEqual( repr(x1), repr(x) )
        self.assertEqual( repr(y1), repr(y) )
        self.assertEqual( repr(z1), repr(z) )
        
        self.assertEqual( get_correlation(x1,y1), r )

        x1, y1, z1 = ar.extract('x4','y4','z4')

        x = ucomplex(1,[10,2,2,10],5)
        y = ucomplex(1-6j,[10,2,2,10],7)
        
        z = result( log( x * y ) )
            
        self.assertEqual( repr(x1), repr(x) )
        self.assertEqual( repr(y1), repr(y) )
        self.assertEqual( repr(z1), repr(z) )

#-----------------------------------------------------
class TestArchivePickleFilev133(unittest.TestCase):

    """
    Use a reference file created using GTC v 1.3.3 
    to make sure we maintain backward compatibility.
    """
    
    def test(self):
        
        fname = 'ref_file_v_1_3_3.gar'
        
        wdir =  os.path.dirname(__file__)
        path = os.path.join(wdir,fname)

        with open(path,'rb') as f:
            ar = persistence.load(f)
            f.close()

        x1, y1, z1 = ar.extract('x','y','z')
        
        self.assertEqual( repr(x1), repr(ureal(1,1)) )
        self.assertEqual( repr(y1), repr(ureal(2,1)) )
        self.assertEqual( repr(z1), repr( ureal(1,1) + ureal(2,1) ) )

        z1 = ar.extract('z1')
        
        self.assertEqual( 
            repr(z1), 
            repr( 
                ureal(1,1,3) + ureal(2,1,4) 
            ) 
        )

        x1, y1, z1 = ar.extract('x2','y2','z2')

        x = ureal(1,1,independent=False)
        y = ureal(2,1,independent=False)
        r = 0.5
        set_correlation(r,x,y)

        self.assertEqual( repr(x1), repr(x) )
        self.assertEqual( repr(y1), repr(y) )
        self.assertEqual( repr(z1), repr(x+y) )

        x1, y1, z1 = ar.extract('x3','y3','z3')

        x,y = multiple_ureal([1,2],[1,1],4)     
        r = 0.5
        set_correlation(r,x,y)
        z = result( x + y )            
        self.assertEqual( repr(x1), repr(x) )
        self.assertEqual( repr(y1), repr(y) )
        self.assertEqual( repr(z1), repr(z) )
        
        self.assertEqual( get_correlation(x1,y1), r )

        x1, y1, z1 = ar.extract('x4','y4','z4')

        x = ucomplex(1,[10,2,2,10],5)
        y = ucomplex(1-6j,[10,2,2,10],7)
        
        z = result( log( x * y ) )
            
        self.assertEqual( repr(x1), repr(x) )
        self.assertEqual( repr(y1), repr(y) )
        self.assertEqual( repr(z1), repr(z) )
        
#============================================================================
if(__name__== '__main__'):

    unittest.main()    # Runs all test methods starting with 'test'


import unittest
import os
import numpy

try:
    from itertools import izip  # Python 2
except ImportError:
    izip = zip

from GTC import *
from GTC.context import Context
from GTC import context
from GTC import persistence
from GTC.vector import is_ordered
from GTC.linear_algebra import uarray, UncertainArray

from testing_tools import *

TOL = 1E-13

#----------------------------------------------------------------------------
# Note these tests simultaneously check `sensitivity` and `u_component` 
#
class TestSensitivity(unittest.TestCase):

    """
    reporting.sensitivity is a function that evaluates
    the linear partial derivatives. 

    Here we can use the u_component calculations to
    test sensitivity because they are independent
    when elementary UNs are concerned.
    """

    def test1(self):
        """Reals
        """
        x1 = ureal(2,1)
        x2 = ureal(5,1)
        x3 = ureal(7,1)

        y = x3 ** x1 * x2

        self.assertTrue(
            equivalent(
                rp.u_component(y,x1),
                rp.sensitivity(y,x1)
        ))
        
        self.assertTrue(
            equivalent(
                rp.u_component(y,x2),
                rp.sensitivity(y,x2)
        ))

        self.assertTrue(
            equivalent(
                rp.u_component(y,x3),
                rp.sensitivity(y,x3)
        ))

        x4 = result( y )
        y = sin(x4)
        # Do the differentiation by hand
        self.assertTrue(
            equivalent(
                math.cos(x4.x),
                rp.sensitivity(y,x4)
        ))


    def test2(self):
        """Complex
        """

        z1 = ucomplex(1+2j,1)
        z2 = ucomplex(2-1.5j,1)
        z3 = ucomplex(.1+1.8j,1)

        z = z1**z2 * z3
        
        self.assertTrue(
            equivalent_sequence(
                rp.u_component(z,z1),
                rp.sensitivity(z,z1)
        ))

        self.assertTrue(
            equivalent_sequence(
                rp.u_component(z,z2),
                rp.sensitivity(z,z2)
        ))

        self.assertTrue(
            equivalent_sequence(
                rp.u_component(z,z3),
                rp.sensitivity(z,z3)
        ))

        z4 = result( z )
        z = sqrt(z4)
        # Do the differentiation by hand
        dz_dz4 = fn.complex_to_seq(
            1.0/(2 * cmath.sqrt(z4.x))
        )

        self.assertTrue(
            equivalent_sequence(
                dz_dz4,
                rp.sensitivity(z,z4)
        ))

    # def test_uarray_singles_ureal(self):
        # """
        # When uarray has just one element 
        
        # """
        # x1 = uarray(ureal(2,1))
        # x2 = uarray(ureal(5,1))
        # x3 = uarray(ureal(7,1))

        # y = x3 ** x1 * x2
        
        # dy_dx = rp.sensitivity(y,x1)
        # uy_x = rp.u_component(y,x1)       
        # self.assertTrue( equivalent(dy_dx,uy_x) ) 

        # dy_dx = rp.sensitivity(y,x2)
        # uy_x = rp.u_component(y,x2)        
        # self.assertTrue( equivalent(dy_dx,uy_x) ) 

        # dy_dx = rp.sensitivity(y,x3)
        # uy_x = rp.u_component(y,x3)        
        # self.assertTrue( equivalent(dy_dx,uy_x) ) 
        
        # x4 = result( y )
        # y = sin(x4)
        # # Do the differentiation by hand
        # self.assertTrue(
            # equivalent(math.cos(x4.x),rp.sensitivity(y,x4))
        # )
        
    # def test_uarray_singles_ucomplex(self):
        # """
        # When uarray has just one element 
        
        # """
        # x1 = uarray(ucomplex(1+2j,1))
        # x2 = uarray(ucomplex(2-1.5j,1))
        # x3 = uarray(ucomplex(.1+1.8j,1))

        # z = x1**x2 * x3
        
        # dz_dx = rp.sensitivity(z,x1)
        # uz_x = rp.u_component(z,x1)       
        # self.assertTrue( equivalent_sequence(dz_dx,uz_x) )
        
        # dz_dx = rp.sensitivity(z,x2)
        # uz_x = rp.u_component(z,x2)       
        # self.assertTrue( equivalent_sequence(dz_dx,uz_x) )
        
        # dz_dx = rp.sensitivity(z,x3)
        # uz_x = rp.u_component(z,x3)       
        # self.assertTrue( equivalent_sequence(dz_dx,uz_x) )
      
        # x4 = result( z )
        # z = sqrt(x4)
        # # Do the differentiation by hand
        # dz_dx4 = fn.complex_to_seq(
            # 1.0/(2 * cmath.sqrt(x4.x))
        # )

        # self.assertTrue(
            # equivalent_sequence(
                # dz_dx4,
                # rp.sensitivity(z,x4)
        # ))
        
    def test_uarray_ureal(self):
        """
        When uarray has more than one element 
        
        """
        x = uarray( [ureal(2,1), ureal(5,1), ureal(7,1) ] )

        y = sin(x)
        
        dy_dx = rp.sensitivity(y,x)
        uy_x = rp.u_component(y,x)    

        self.assertTrue( isinstance(dy_dx,UncertainArray) )
        self.assertTrue( isinstance(uy_x,UncertainArray) )
        
        for dy_dx_i, uy_x_i in zip( dy_dx, uy_x ):
            self.assertTrue( equivalent(dy_dx_i,uy_x_i) ) 

        # Do the differentiation by hand
        dy_dx_calc = cos( value(x) )
        for dy_dx_i, dy_dx_calc_i in zip( dy_dx, dy_dx_calc ):
            self.assertTrue( equivalent(dy_dx_i,dy_dx_calc_i) ) 
        
    def test_uarray_ucomplex(self):
        """
        When uarray has more than one element 
        
        """
        x = uarray( [ucomplex(1+2j,1),ucomplex(2-1.5j,1),ucomplex(.1+1.8j,1)] )

        z = sin(x)
        
        dz_dx = rp.sensitivity(z,x)
        uz_x = rp.u_component(z,x)       
        for dz_dx_i, uz_x_i in zip( dz_dx, uz_x ):
            self.assertTrue( equivalent_sequence(dz_dx_i,uz_x_i) ) 

        # Do the differentiation by hand
        dx_dx_calc = cos( value(x) )
        for dz_dx_i, uz_x_i in zip( dz_dx, uz_x ):
            self.assertTrue( equivalent_sequence(dz_dx_i,uz_x_i) ) 

    def test_uarray_mixed(self):
        """
        When uarray has more than one element of different types
        
        """
        x = uarray( [ ureal(2,1), ucomplex(2-1.5j,1) ] )
        y = sin(x)
        
        dy_dx = rp.sensitivity(y,x)
        uy_x = rp.u_component(y,x)    

        self.assertTrue( isinstance(dy_dx,UncertainArray) )
        self.assertTrue( isinstance(uy_x,UncertainArray) )
        
        self.assertTrue( equivalent(dy_dx[0],uy_x[0]) ) 
        self.assertTrue( equivalent_sequence(dy_dx[1],uy_x[1]) ) 

        # Do the differentiation by hand
        dy_dx_calc = cos( value(x) ) 
        self.assertTrue( equivalent(dy_dx[0],dy_dx_calc[0]) )
        self.assertTrue( equivalent(dy_dx[0],dy_dx_calc[0]) )
        self.assertTrue( 
            equivalent_sequence(
                dy_dx[1],
                fn.complex_to_seq( dy_dx_calc[1])
            ) 
        )             
        
    def test_intermediate_ureal(self):
        """
        When uarray has more than one element of different types
        
        """
        x1 = uarray( [ureal(2,1), ureal(1.5,1)] )
        x2 = uarray( [ureal(5,1), ureal(.2,1)] )
        x3 = uarray( [ureal(7,1), ureal(3.2,1)] )
        
        x4 = result(x3 ** x1 * x2)        
        y = log10(x4) 
        
        dy_dx = rp.sensitivity(y,x4)
        uy_x = rp.u_component(y,x4)       
        self.assertTrue( equivalent_sequence(dy_dx,uy_x/x4.u) ) 
          
        x4 = result(x3 ** x1 * x2, label='x4')        
        y = log10(x4) 
     
        for i,x4_i in enumerate(x4):
            self.assertEqual('x4[{0}]'.format(i),x4_i.label)
            
        dy_dx = rp.sensitivity(y,x4)
        uy_x = rp.u_component(y,x4)       
        self.assertTrue( equivalent_sequence(dy_dx,uy_x/x4.u) ) 
  
        label = [ "x4[{}]".format(i) for i in range(x4.size) ]
        x4 = result(x3 ** x1 * x2, label=label)        
        y = log10(x4) 
     
        for i,x4_i in enumerate(x4):
            self.assertEqual('x4[{0}]'.format(i),x4_i.label)
 
    def test_intermediate_ucomplex(self):
        """
        When uarray has more than one element of different types
        
        """
        z1 = uarray( [ucomplex(1+2j,1), ucomplex(1.5+2.1j,1)] )
        z2 = uarray( [ucomplex(2-1.5j,1), ucomplex(.5-0.5j,1)] )
        z3 = uarray( [ucomplex(.1+1.8j,1), ucomplex(.4+1.1j,1)] )
        
        z4 = result(z1**z2 * z3)        
        z = log(z4) 
       
        for z_i, z4_i in zip(z,z4):
        
            uz_i_z4_i = rp.u_component(z_i,z4_i)
            dz_i_dz4_i = rp.sensitivity(z_i,z4_i)
            
            self.assertTrue(
                equivalent(dz_i_dz4_i.rr,uz_i_z4_i.rr/z4_i.real.u)
            )
            self.assertTrue(
                equivalent(dz_i_dz4_i.ir,uz_i_z4_i.ir/z4_i.real.u)
            )
            self.assertTrue(
                equivalent(dz_i_dz4_i.ri,uz_i_z4_i.ri/z4_i.imag.u)
            )
            self.assertTrue(
                equivalent(dz_i_dz4_i.ii,uz_i_z4_i.ii/z4_i.imag.u)
            )
        
        z4 = result(z1**z2 * z3, label='z4')        
        z = log(z4) 
     
        for i,z4_i in enumerate(z4):
            self.assertEqual('z4[{0}]'.format(i),z4_i.label)
            self.assertEqual('z4[{0}]_re'.format(i),z4_i.real.label)
            self.assertEqual('z4[{0}]_im'.format(i),z4_i.imag.label)
            
        for z_i, z4_i in zip(z,z4):
        
            uz_i_z4_i = rp.u_component(z_i,z4_i)
            dz_i_dz4_i = rp.sensitivity(z_i,z4_i)
            
            self.assertTrue(
                equivalent(dz_i_dz4_i.rr,uz_i_z4_i.rr/z4_i.real.u)
            )
            self.assertTrue(
                equivalent(dz_i_dz4_i.ir,uz_i_z4_i.ir/z4_i.real.u)
            )
            self.assertTrue(
                equivalent(dz_i_dz4_i.ri,uz_i_z4_i.ri/z4_i.imag.u)
            )
            self.assertTrue(
                equivalent(dz_i_dz4_i.ii,uz_i_z4_i.ii/z4_i.imag.u)
            )

        label = [ "z4[{}]".format(i) for i in range(z4.size) ]
        z4 = result(z1**z2 * z3, label=label)        
        z = log(z4) 
     
        for i,z4_i in enumerate(z4):
            self.assertEqual('z4[{0}]'.format(i),z4_i.label)
            self.assertEqual('z4[{0}]_re'.format(i),z4_i.real.label)
            self.assertEqual('z4[{0}]_im'.format(i),z4_i.imag.label) 
#=====================================================
if(__name__== '__main__'):

    unittest.main()    # Runs all test methods starting with 'test'


"""
Sample estimates
----------------
    *   :func:`estimate` returns an uncertain number defined from
        the statistics of a sample of data.
    *   :func:`multi_estimate_real` returns a sequence of related 
        uncertain real numbers defined from the multivariate statistics 
        calculated from a sample of data. 
    *   :func:`multi_estimate_complex` returns a sequence of related uncertain
        complex numbers defined from the multivariate statistics of a sample of data. 
    *   :func:`estimate_digitized` returns an uncertain number for 
        the mean of a sample of digitized data.
    *   :func:`mean` returns the mean of a sample of data.
    *   :func:`standard_uncertainty` evaluates the standard 
        uncertainty associated with the sample mean.
    *   :func:`standard_deviation` evaluates the standard 
        deviation of a sample of data.
    *   :func:`variance_covariance_complex` evaluates the variance
        and covariance associated with the mean real component 
        and mean imaginary component of the data.
    
Least squares regression
------------------------
    *   :func:`line_fit` performs an ordinary least-squares straight 
        line fit to a sample of data.  
    *   :func:`line_fit_wls` performs a weighted least-squares straight 
        line fit to a sample of data. The weights are assumed to be exact.
    *   :func:`line_fit_rwls` performs a weighted least-squares  
        straight line fit to a sample of data. The weights
        are only assumed normalise the variability of observations.
    *   :func:`line_fit_wtls` performs a weighted total least-squares straight 
        line fit to a sample of data.   
 
Merging uncertain components
----------------------------
    *   :func:`merge` combines results from a type-A and type-B analysis 
        of the same data. 
 
.. note::

    Many functions in :mod:`type_a` treat  data as pure numbers. 
    Sequences of uncertain numbers can be passed to these 
    functions, but only the uncertain-number values will be used.
    
    :func:`merge` is provided so that the results of 
    type-A and type-B analyses on the same data can be 
    combined. 

Module contents
---------------

"""
from __future__ import division

import sys
import math
import numbers
import numpy as np
from functools import reduce

try:
    from itertools import izip  # Python 2
    from collections import Iterable
except ImportError:
    izip = zip
    xrange = range
    from collections.abc import Iterable

from GTC import (
    inf, 
    function,
    is_sequence
)
from GTC.lib import (
    UncertainReal, 
    UncertainComplex,
    set_correlation_real,
    real_ensemble,
    complex_ensemble,
    append_real_ensemble,
    value,
    value_seq
)
ureal = UncertainReal._elementary
ucomplex = UncertainComplex._elementary
result = lambda un,label: un._intermediate(label)

from GTC import type_b
from GTC.type_b import (
    LineFit, LineFitWTLS,
)

from GTC.named_tuples import (
    VarianceCovariance,
    StandardUncertainty,
    StandardDeviation,
    InterceptSlope
)

EPSILON = sys.float_info.epsilon 
HALF_PI = math.pi/2.0

__all__ = (
    'estimate',
    'estimate_digitized',
    'multi_estimate_real',
    'multi_estimate_complex',
    'mean',
    'standard_deviation',
    'standard_uncertainty',
    'variance_covariance_complex',
    'line_fit', 'line_fit_wls', 'line_fit_rwls', 'line_fit_wtls',
    'LineFitOLS','LineFitRWLS','LineFitWLS','LineFitWTLS',
    'merge',
)

#-----------------------------------------------------------------------------------------
#
class LineFitOLS(LineFit):
    
    """
    Class to hold the results of an ordinary least-squares regression to data.

    It can also be used to apply the results of a regression analysis. 
    
    .. versionadded:: 1.2
    """
    
    def __init__(self,a,b,ssr,N):
        LineFit.__init__(self,a,b,ssr,N)
            
    def __str__(self):
        header = '''
Ordinary Least-Squares Results:
'''
        return header + LineFit.__str__(self)

    def x_from_y(self,yseq,x_label=None,y_label=None):
        """Estimate the stimulus ``x`` corresponding to the responses in ``yseq``

        :arg yseq: a sequence of ``y`` observations 
        :arg x_label: a label for the return uncertain number `x` 
        :arg y_label: a label for the estimate of `y` based on ``yseq`` 

        .. note::
            When ``x_label`` is defined, the uncertain number returned will be 
            declared an intermediate result (using :func:`~.result`)

        **Example** ::
        
            >>> x_data = [0.1, 0.1, 0.1, 0.3, 0.3, 0.3, 0.5, 0.5, 0.5,
            ...                 0.7, 0.7, 0.7, 0.9, 0.9, 0.9]
            >>> y_data = [0.028, 0.029, 0.029, 0.084, 0.083, 0.081, 0.135, 0.131,
            ...                 0.133, 0.180, 0.181, 0.183, 0.215, 0.230, 0.216]

            >>> fit = type_a.line_fit(x_data,y_data)
            
            >>> x0 = fit.x_from_y( [0.0712, 0.0716] )            
            >>> x0
            ureal(0.2601659751037...,0.01784461112558...,13.0)

        """
        df = self._N - 2       
        a, b = self._a_b

        p = len(yseq)
        y = math.fsum( yseq ) / p
        
        y = ureal(
            y,
            math.sqrt( self._ssr/df/p ),
            df,
            label=y_label,
            independent=False
        )  

        append_real_ensemble(a,y)
        
        if x_label is None:
            x = (y - a)/b
        else:
            x = result( (y - a)/b, label=x_label )

        return x
        
    def y_from_x(self,x,s_label=None,y_label=None):
        """Return an uncertain number ``y`` that predicts the response to ``x``

        :arg x: a real number, or an uncertain real number
        :arg s_label: a label for an elementary uncertain number associated with observation variability  
        :arg y_label: a label for the return uncertain number `y` 

        This is a prediction of a single future response ``y`` to a stimulus ``x``
        
        The variability in observations is based on residuals obtained during regression.
        
        An uncertain real number can be used for ``x``, in which
        case the associated uncertainty will also be propagated into ``y``.

        .. note::
            When ``y_label`` is defined, the uncertain number returned will be 
            declared an intermediate result (using :func:`~.result`)
        
        """
        a, b = self._a_b   
        df = self._N - 2
        
        u = math.sqrt( self._ssr/df )
        
        noise = ureal(
            0,u,df,label=s_label,independent=False
        )
        
        append_real_ensemble(a,noise)
                  
        if y_label is None:
            y = a + b*x + noise
        else:
            y = result( a + b*x + noise, label=y_label )
        
        return y

#-----------------------------------------------------------------------------------------
#
class LineFitRWLS(LineFit):
    
    """
    Class to hold the results of a relative weighted least-squares regression.
    The weights provided normalise the variability of observations.
    
    .. versionadded:: 1.2
    """
    
    def __init__(self,a,b,ssr,N):
        LineFit.__init__(self,a,b,ssr,N)

    def __str__(self):
        header = '''
Relative Weighted Least-Squares Results:
'''
        return header + LineFit.__str__(self)

    def x_from_y(self,yseq,s_y,x_label=None,y_label=None):
        """Estimate the stimulus ``x`` corresponding to the responses in ``yseq``

        :arg yseq: a sequence of further observations of ``y``
        :arg s_y: a scale factor for the uncertainty of the ``yseq``
        :arg x_label: a label for the return uncertain number `x` 
        :arg y_label: a label for the estimate of `y` based on ``yseq``

        .. note::
            When ``x_label`` is defined, the uncertain number returned will be 
            declared an intermediate result (using :func:`~.result`)

        """
        df = self._N - 2       
        a, b = self._a_b
        
        p = len(yseq)
        y = math.fsum( yseq ) / p
        
        y = ureal(
            y,
            s_y * math.sqrt( self._ssr/df/p ),
            df,
            label=y_label,
            independent=False
        )            

        append_real_ensemble(a,y)
        
        if x_label is None:
            x = (y - a)/b
        else:
            x = result( (y - a)/b, label=x_label )

        return x

    def y_from_x(self,x,s_y,s_label=None,y_label=None):
        """Return an uncertain number ``y`` that predicts the response to ``x``

        :arg x: a real number, or an uncertain real number
        :arg s_y: a scale factor for the response uncertainty
        :arg s_label: a label for an elementary uncertain number associated with observation variability  
        :arg y_label: a label for the return uncertain number `y` 

        Returns a single future response ``y`` predicted for a stimulus ``x``.

        Because there is different variability in 
        the response to different stimuli, the
        scale factor ``s_y`` is required. It is assumed 
        that the standard deviation in the ``y`` value is 
        proportional to ``s_y``.
        
        An uncertain real number can be used for ``x``, in which
        case the associated uncertainty is also propagated into ``y``.
        
        .. note::
            When ``y_label`` is defined, the uncertain number returned will be 
            declared an intermediate result (using :func:`~.result`)

        """
        a, b = self._a_b   
        df = self._N - 2
        
        u = math.sqrt( s_y*self._ssr/df )
        
        noise = ureal(0,u,df,label=s_label)

        append_real_ensemble(a,noise)
                  
        if y_label is None:
            y = a + b*x + noise
        else:
            y = result( a + b*x + noise, label=y_label )
        
        return y
        
#-----------------------------------------------------------------------------------------
#
class LineFitWLS(LineFit):
    
    """
    Class to hold the results of a weighted least-squares regression.
    The weight factors provided are assumed to correspond exactly to the
    variability of observations.
    
    .. versionadded:: 1.2
    """
    
    def __init__(self,a,b,ssr,N):
        LineFit.__init__(self,a,b,ssr,N)

    def __str__(self):
        header = '''
Weighted Least-Squares Results:
'''
        return header + LineFit.__str__(self)

    def x_from_y(self,yseq,u_yseq,x_label=None,y_label=None):
        """Estimate the stimulus ``x`` corresponding to the responses in ``yseq``

        :arg yseq: a sequence of further observations of ``y``
        :arg u_yseq: the standard uncertainty of the ``yseq`` data
        :arg x_label: a label for the return uncertain number `x` 
        :arg y_label: a label for the estimate of `y` based on ``yseq``

        The variations in ``yseq`` values are assumed to result from 
        independent random effects.
        
        .. note::
            When ``x_label`` is defined, the uncertain number returned will be 
            declared an intermediate result (using :func:`~.result`)

        """
        a, b = self._a_b
        
        p = len(yseq)
        y = math.fsum( yseq ) / p
        
        y = ureal(
            y,
            u_yseq / math.sqrt( p ),
            inf,
            label=y_label,
            independent=False
        )            

        append_real_ensemble(a,y)
        
        if x_label is None:
            x = (y - a)/b
        else:
            x = result( (y - a)/b, label=x_label )

        return x

    def y_from_x(self,x,s_y,s_label=None,y_label=None):
        """Return an uncertain number ``y`` that predicts the response to ``x``

        :arg x: a real number, or an uncertain real number
        :arg s_y: response variability uncertainty
        :arg s_label: a label for an elementary uncertain number associated with response variability  
        :arg y_label: a label for the return uncertain number `y` 

        Returns a single future response ``y`` predicted for a stimulus ``x``.

        The standard uncertainty ``s_y`` is used to create an additive
        component of uncertainty associated with variability in the ``y`` value.
        
        An uncertain real number can be used for ``x``, in which
        case the associated uncertainty is also propagated into ``y``.
        
        .. note::
            When ``y_label`` is defined, the uncertain number returned will be 
            declared an intermediate result (using :func:`~.result`)

        """
        a, b = self._a_b   
        df = self.N - 2
        
        noise = ureal(0,s_y,df,label=s_label)

        append_real_ensemble(a,noise)
                  
        if y_label is None:
            y = a + b*x + noise
        else:
            y = result( a + b*x + noise, label=y_label )
        
        return y
        
#-----------------------------------------------------------------------------------------
def line_fit(x,y,label=None):
    """Return a least-squares straight-line fit to the data
     
    .. versionadded:: 1.2
    
    :arg x:     sequence of stimulus data (independent-variable)  
    :arg y:     sequence of response data (dependent-variable)  
    :arg label: suffix to label the uncertain numbers `a` and `b`

    :returns:   an object containing regression results
    :rtype:     :class:`.LineFitOLS`

    Performs an ordinary least-squares regression of ``y`` to ``x``.
        
    **Example**::
    
        >>> x = [1,2,3,4,5,6,7,8,9]
        >>> y = [15.6,17.5,36.6,43.8,58.2,61.6,64.2,70.4,98.8]
        >>> result = type_a.line_fit(x,y)
        >>> a,b = result.a_b
        >>> a
        ureal(4.8138888888888...,4.8862063121833...,7)
        >>> b
        ureal(9.4083333333333...,0.8683016476563...,7)

        >>> y_p = a + b*5.5
        >>> dof(y_p)
        7.0
            
    """
    N = len(x)
    
    x = value_seq(x)
    y = value_seq(y)

    df = N-2

    S = N
    S_x = math.fsum( x )
    S_y = math.fsum( y )
        
    k = S_x / S
    t = [ (x_i - k) for x_i in x ]

    S_tt =  math.fsum( t_i*t_i for t_i in t )

    b_ =  math.fsum( t_i*y_i/S_tt for t_i,y_i in izip(t,y) )
    a_ = (S_y - b_*S_x)/S

    siga = math.sqrt( (1.0 + S_x*S_x/(S*S_tt))/S )
    sigb = math.sqrt( 1.0/S_tt )
    r_ab = -S_x/(S*S_tt*siga*sigb)
    
    # Sum of squared residuals needed to correctly calculate parameter uncertainties
    f = lambda x_i,y_i: (y_i - a_ - b_*x_i)**2 
    ssr =  math.fsum( f(x_i,y_i) for x_i,y_i in izip(x,y) )

    data_u = math.sqrt( ssr/df )
    siga *= data_u
    sigb *= data_u
            
    a = ureal(
        a_,
        siga,
        df=df,
        label='a_{}'.format(label) if label is not None else None,
        independent=False
    )
    b = ureal(
        b_,
        sigb,
        df=df,
        label='b_{}'.format(label) if label is not None else None,
        independent=False
    )
    
    real_ensemble( (a,b), df )
    a.set_correlation(r_ab,b)

    return LineFitOLS(a,b,ssr,N)

#-----------------------------------------------------------------------------------------
def _line_fit_wls(x,y,u_y):
    """Utility function
    
    NB client must supply sequences `x` and `y` of pure numbers
    """
    N = len(x)

    v = [ u_y_i*u_y_i for u_y_i in u_y ]
    S =  math.fsum( 1.0/v_i for v_i in v)

    S_x =  math.fsum( x_i/v_i for x_i,v_i in izip(x,v) )
    S_y =  math.fsum( y_i/v_i for y_i,v_i in izip(y,v) )

    k = S_x / S
    t = [ (x_i - k)/u_y_i for x_i,u_y_i in izip(x,u_y) ]

    S_tt =  math.fsum( t_i*t_i for t_i in t )

    b_ =  math.fsum( t_i*y_i/u_y_i/S_tt for t_i,y_i,u_y_i in izip(t,y,u_y) )
    a_ = (S_y - b_*S_x)/S

    siga = math.sqrt( (1.0 + S_x*S_x/(S*S_tt))/S )
    sigb = math.sqrt( 1.0/S_tt )
    r_ab = -S_x/(S*S_tt*siga*sigb)
    
    f = lambda x_i,y_i,u_y_i: ((y_i - a_ - b_*x_i)/u_y_i)**2 
    ssr =  math.fsum( f(x_i,y_i,u_y_i) for x_i,y_i,u_y_i in izip(x,y,u_y) )

    return a_,b_,siga,sigb,r_ab,ssr,N

#-----------------------------------------------------------------------------------------
def line_fit_wls(x,y,u_y,label=None):
    """Return a weighted least-squares straight-line fit
    
    .. versionadded:: 1.2
     
    :arg x:     sequence of stimulus data (independent-variable)  
    :arg y:     sequence of response data (dependent-variable)  
    :arg u_y:   sequence of uncertainties in the response data 
    :arg label: suffix to label the uncertain numbers `a` and `b`

    :returns:   an object containing regression results
    :rtype:     :class:`.LineFitWLS`

    **Example**::
    
        >>> x = [1,2,3,4,5,6]
        >>> y = [3.2, 4.3, 7.6, 8.6, 11.7, 12.8]
        >>> u_y = [0.5,0.5,0.5,1.0,1.0,1.0]
        
        >>> fit = type_a.line_fit_wls(x,y,u_y)
        >>> fit.a_b     
         InterceptSlope(a=ureal(0.8852320675105...,0.5297081435088...,inf),
         b=ureal(2.056962025316...,0.177892016741...,inf))
        
    """
    x = value_seq(x)
    y = value_seq(y)

    a_,b_,siga,sigb,r_ab,ssr,N = _line_fit_wls(x,y,u_y)
    
    a = ureal(
        a_,
        siga,
        inf,
        label='a_{}'.format(label) if label is not None else None,
        independent=False
    )
    b = ureal(
        b_,
        sigb,
        inf,
        label='b_{}'.format(label) if label is not None else None,
        independent=False
    )
    
    a.set_correlation(r_ab,b)

    return LineFitWLS(a,b,ssr,N)

#-----------------------------------------------------------------------------------------
def line_fit_rwls(x,y,s_y,label=None):
    """Return a relative weighted least-squares straight-line fit
    
    .. versionadded:: 1.2
    
    The ``s_y`` values are used to scale variability in the ``y`` data.
    It is assumed that the standard deviation of each ``y`` value is 
    proportional to the corresponding ``s_y`` scale factor.
    The unknown common factor in the uncertainties is estimated from the residuals.
    
    :arg x:     sequence of stimulus data (independent-variable)  
    :arg y:     sequence of response data (dependent-variable)  
    :arg s_y:   sequence of scale factors
    :arg label: suffix to label the uncertain numbers `a` and `b`

    :returns:   an object containing regression results
    :rtype:     :class:`.LineFitRWLS`

    **Example**::

        >>> x = [1,2,3,4,5,6]
        >>> y = [3.014,5.225,7.004,9.061,11.201,12.762]
        >>> s_y = [0.2,0.2,0.2,0.4,0.4,0.4]
        >>> fit = type_a.line_fit_rwls(x,y,s_y)
        >>> a, b = fit.a_b
        >>>
        >>> print(fit)
        <BLANKLINE>
        Relative Weighted Least-Squares Results:
        <BLANKLINE>
          Intercept: 1.14(12)
          Slope: 1.973(41)
          Correlation: -0.87
          Sum of the squared residuals: 1.3395217958...
          Number of points: 6
        <BLANKLINE>
          
    """
    N = len(x)
    if N < 3:
        raise RuntimeError(
            "At least three data points are required, got {}".format(N)
        )
        
    x = value_seq(x)
    y = value_seq(y)
    
    a_,b_,siga,sigb,r_ab,ssr,N = _line_fit_wls(x,y,s_y)

    df = N-2
    sigma_hat = math.sqrt(ssr/df)
    siga *= sigma_hat
    sigb *= sigma_hat
    
    a = ureal(
        a_,
        siga,
        df,
        label='a_{}'.format(label) if label is not None else None,
        independent=False
    )
    b = ureal(
        b_,
        sigb,
        df,
        label='b_{}'.format(label) if label is not None else None,
        independent=False
    )
    
    real_ensemble( (a,b), df )
    a.set_correlation(r_ab,b)

    return LineFitRWLS(a,b,ssr,N)
    
#--------------------------------------------------------------------
#
def line_fit_wtls(x,y,u_x,u_y,a0_b0=None,r_xy=None,label=None):
    """Return a total least-squares straight-line fit 
    
    .. versionadded:: 1.2

    :arg x:     sequence of independent-variable data
    :arg y:     sequence of dependent-variable data 
    :arg u_x:   sequence of uncertainties in ``x``
    :arg u_y:   sequence of uncertainties in ``y``
    :arg a0_b0: a pair of initial estimates for the intercept and slope
    :arg r_xy:  correlation between x-y pairs
    :arg label: suffix labeling the uncertain numbers `a` and `b`

    :returns:   an object containing the fitting results
    :rtype:     :class:`.LineFitWTLS`

    The optional argument ``a_b`` can be used to provide a pair 
    of initial estimates for the intercept and slope. 

    Based on paper by M Krystek and M Anton,
    *Meas. Sci. Technol.* **22** (2011) 035101 (9pp)
    
    **Example**::

        # Pearson-York test data see, e.g., 
        # Lybanon, M. in Am. J. Phys 52 (1) 1984 
        >>> x=[0.0,0.9,1.8,2.6,3.3,4.4,5.2,6.1,6.5,7.4]
        >>> wx=[1000.0,1000.0,500.0,800.0,200.0,80.0,60.0,20.0,1.8,1.0]

        >>> y=[5.9,5.4,4.4,4.6,3.5,3.7,2.8,2.8,2.4,1.5]
        >>> wy=[1.0,1.8,4.0,8.0,20.0,20.0,70.0,70.0,100.0,500.0]

        # standard uncertainties required for weighting
        >>> ux=[1./math.sqrt(wx_i) for wx_i in wx ]
        >>> uy=[1./math.sqrt(wy_i) for wy_i in wy ]

        >>> result = ta.line_fit_wtls(x,y,ux,uy)
        >>> intercept, slope = result.a_b
        >>> intercept
        ureal(5.47991018...,0.29193349...,8)
        >>> slope
        ureal(-0.48053339...,0.057616740...,8)
    
    """
    independent = r_xy is None

    x_u = [ ureal( value(x_i),u_i,inf,None,independent=independent)
        for x_i, u_i in izip(x,u_x)
    ]
    y_u = [ ureal( value(y_i),u_i,inf,None,independent=independent)
        for y_i, u_i in izip(y,u_y)
    ]
    if not independent:
        for x_i,y_i,r_i in izip(x_u,y_u,r_xy):
            x_i.set_correlation(r_i,y_i)

    result = type_b.line_fit_wtls(x_u,y_u,a_b=a0_b0)

    a, b = result.a_b
    N = result.N
    ssr = result.ssr
    r_ab = a.get_correlation(b)
    
    df = N-2
    
    a = ureal(
        a.x,
        a.u,
        df,
        label='a_{}'.format(label) if label is not None else None,
        independent=False
    )
    b = ureal(
        b.x,
        b.u,
        df,
        label='b_{}'.format(label) if label is not None else None,
        independent=False
    )

    real_ensemble( (a,b), df )
    a.set_correlation(r_ab,b)

    return LineFitWTLS(a,b,ssr,N)
    
#-----------------------------------------------------------------------------------------
def estimate_digitized(seq,delta,label=None,truncate=False):
    """
    Return an uncertain number for the mean of digitized data in ``seq``

    :arg seq: data
    :type seq: float, :class:`~lib.UncertainReal` or :class:`~lib.UncertainComplex`
    :arg float delta: digitization step size 
    :arg str label: label for uncertain number returned
    :arg bool truncate: if ``True``, truncation, rather than rounding, is assumed
    :rtype: :class:`~lib.UncertainReal` or :class:`~lib.UncertainComplex`

    A sequence of data that has been formatted with fixed precision  
    can completely conceal a small amount of variability in the original
    values, or merely obscure that variability.  
    
    This function recognises the possible interaction between truncation, or rounding,
    errors and random errors in the underlying data. The function 
    evaluates the mean of the data and evaluates the uncertainty 
    in this mean.   
        
    Set the argument ``truncate`` to ``True`` 
    if data have been truncated, instead of rounded.
    
    See reference: R Willink, *Metrologia*, **44** (2007) 73-81

    **Examples**::
    
        # LSD = 0.0001, data varies between -0.0055 and -0.0057
        >>> seq = (-0.0056,-0.0055,-0.0056,-0.0056,-0.0056, 
        ...      -0.0057,-0.0057,-0.0056,-0.0056,-0.0057,-0.0057)
        >>> type_a.estimate_digitized(seq,0.0001)
        ureal(-0.005627272727272...,1.9497827808661...e-05,10)

        # LSD = 0.0001, data varies between -0.0056 and -0.0057
        >>> seq = (-0.0056,-0.0056,-0.0056,-0.0056,-0.0056,
        ... -0.0057,-0.0057,-0.0056,-0.0056,-0.0057,-0.0057)
        >>> type_a.estimate_digitized(seq,0.0001)
        ureal(-0.005636363636363...,1.52120004824377...e-05,10)

        # LSD = 0.0001, no spread in data values
        >>> seq = (-0.0056,-0.0056,-0.0056,-0.0056,-0.0056,
        ... -0.0056,-0.0056,-0.0056,-0.0056,-0.0056,-0.0056)
        >>> type_a.estimate_digitized(seq,0.0001)
        ureal(-0.0056,2.8867513459481...e-05,10)
        
        # LSD = 0.0001, no spread in data values, fewer points
        >>> seq = (-0.0056,-0.0056,-0.0056)
        >>> type_a.estimate_digitized(seq,0.0001)
        ureal(-0.0056,3.2914029430219...e-05,2)
        
    """
    N = len(seq)
    if N < 2:
        raise RuntimeError(
            "digitized data sequence must have more than one element"
        )

    seq = value_seq(seq)
    
    x_max = max(seq)
    x_min = min(seq)
    
    mu = mean(seq)
        
    if x_max == x_min:
        # No scatter in the data
        if N == 2:
            root_c_12 = math.sqrt(6.4/12.0)
        elif N == 3:
            root_c_12 = math.sqrt(1.3/12.0)
        elif N >= 4:
            root_c_12 = math.sqrt(1.0/12.0)
        else:
            assert False,"should not occur"
            
        u = root_c_12*delta
    else:
        accum = lambda psum,x: psum + (x-mu)**2
        var = reduce(accum, seq, 0.0) / (N - 1)

        if abs(x_max - x_min - delta) < 10*sys.float_info.epsilon:
            # Scatter is LSD only
            x_mid = (x_max + x_min)/2.0
            u = math.sqrt(
                max(var/N,(x_mid - mu)**2/3.0)
            )
        else:
            u = math.sqrt(var/N)

    if truncate:
        mu += delta/2.0
        
    return ureal(mu,u,N-1,label,independent=True)
    
#-----------------------------------------------------------------------------------------
def estimate(seq,label=None):
    """Return an uncertain number for the mean of the data in ``seq``

    :arg seq:   a sequence of data
    :arg str label: a label for the returned uncertain number
    
    :rtype:   :class:`~lib.UncertainReal` or :class:`~lib.UncertainComplex`
                
    The elements of ``seq`` may be real numbers, complex numbers, or
    uncertain real or complex numbers. Note that only the value of uncertain 
    numbers will be used.
    
    The function returns an :class:`~lib.UncertainReal` when 
    the mean of the data is real, and an :class:`~lib.UncertainComplex` 
    when the mean of the data is complex.

    In a type-A evaluation, the sample mean provides an estimate of the  
    quantity of interest. The uncertainty in this estimate 
    is the standard deviation of the sample mean (or the  
    sample covariance of the mean, in the complex case).    

    **Examples**::

        >>> data = range(15)
        >>> type_a.estimate(data)
        ureal(7.0,1.1547005383792515,14)
        
        >>> data = [(0.91518731126816899+1.5213442955575518j),
        ... (0.96572684493613492-0.18547192979059401j),
        ... (0.23216598132006649+1.6951311687588568j),
        ... (2.1642786101267397+2.2024333895672563j),
        ... (1.1812532664590505+0.59062101107787357j),
        ... (1.2259264339405165+1.1499373179910186j),
        ... (-0.99422341300318684+1.7359338393131392j),
        ... (1.2122867690240853+0.32535154897909946j),
        ... (2.0122536479379196-0.23283009302603963j),
        ... (1.6770229536619197+0.77195994890476838j)]

        >>> type_a.estimate(data)
        ucomplex((1.059187840567141+0.9574410497332932j), u=[0.28881665310241805,0.2655555630050262], r=-0.3137404512459525, df=9)

    """
    df = len(seq)-1
    
    seq = value_seq(seq)
    
    mu = mean(seq)
    
    if isinstance(mu,complex):
        u,r = standard_uncertainty(seq,mu)
        return ucomplex(
            mu,u[0],u[1],r,df,
            label,
            independent = (r == 0.0)
        )
       
    else:
        u = standard_uncertainty(seq,mu)
        return ureal(
            mu,u,df,label,independent=True
        )

#-----------------------------------------------------------------------------------------
def mean(seq,*args,**kwargs):
    """Return the arithmetic mean of data in ``seq``

    :arg seq: a sequence, :class:`~numpy.ndarray`, or iterable, of numbers or uncertain numbers
    :arg args: optional arguments when ``seq`` is an :class:`~numpy.ndarray`
    :arg kwargs: optional keyword arguments when ``seq`` is an :class:`~numpy.ndarray`
    
    If ``seq`` contains real or uncertain real numbers,
    a real number is returned.

    If ``seq`` contains complex or uncertain complex
    numbers, a complex number is returned.
    
    **Example**::

        >>> data = range(15)
        >>> type_a.mean(data)
        7.0
            
    """
    return value( type_b.mean(seq,*args,**kwargs) )
    
#-----------------------------------------------------------------------------------------
def standard_deviation(seq,mu=None):
    """Return the sample standard deviation
    
    :arg seq: sequence of data
    :arg mu: the arithmetic mean of ``seq``
        
    If ``seq`` contains real or uncertain real numbers, 
    the sample standard deviation is returned.
    
    If ``seq`` contains complex or uncertain complex
    numbers, the standard deviation in the real and
    imaginary components is evaluated, as well as
    the correlation coefficient between the components.
    The results are returned in a pair of objects: a
    :obj:`~named_tuples.StandardDeviation` namedtuple 
    and a correlation coefficient. 

    Only the values of uncertain numbers are used in calculations. 
    
    **Examples**::

        >>> data = range(15)
        >>> type_a.standard_deviation(data)
        4.47213595499958

        >>> data = [(0.91518731126816899+1.5213442955575518j),
        ... (0.96572684493613492-0.18547192979059401j),
        ... (0.23216598132006649+1.6951311687588568j),
        ... (2.1642786101267397+2.2024333895672563j),
        ... (1.1812532664590505+0.59062101107787357j),
        ... (1.2259264339405165+1.1499373179910186j),
        ... (-0.99422341300318684+1.7359338393131392j),
        ... (1.2122867690240853+0.32535154897909946j),
        ... (2.0122536479379196-0.23283009302603963j),
        ... (1.6770229536619197+0.77195994890476838j)]
        >>> sd,r = type_a.standard_deviation(data)
        >>> sd
        StandardDeviation(real=0.913318449990377, imag=0.8397604244242309)
        >>> r
        -0.31374045124595246
        
    """
    N = len(seq)
    
    if mu is None:
        mu = mean(seq)

    # `type_a.mean` returns either a real or complex
    if isinstance(mu,numbers.Real):
        accum = lambda psum,x: psum + (value(x)-mu)**2
        variance = reduce(accum, seq, 0.0) / (N - 1)
        
        return math.sqrt( variance )
        
    elif isinstance(mu,numbers.Complex):
        cv_11,cv_12,cv_12,cv_22 = variance_covariance_complex(seq,mu)

        sd_re = math.sqrt(cv_11)
        sd_im = math.sqrt(cv_22)

        den = sd_re * sd_im
        
        if den == 0.0: 
            if cv_12 != 0.0 :
                raise RuntimeError(
                    "numerical instability in covariance calculation"
                )
            else:
                r = 0.0
        else:
            r = cv_12 / den
            
        return StandardDeviation(sd_re,sd_im), r
        
    else:
        raise RuntimeError(
            "unexpected type for mean value: {!r}".format(mu)
        )

#-----------------------------------------------------------------------------------------
def standard_uncertainty(seq,mu=None):
    """Return the standard uncertainty associated with the sample mean

    :arg seq: sequence of data
    :arg mu: the arithmetic mean of ``seq``
    
    :rtype: float or :obj:`~named_tuples.StandardUncertainty`
    
    If ``seq`` contains real or uncertain real numbers,
    the standard uncertainty of the sample mean 
    is returned.

    If ``seq`` contains complex or uncertain complex
    numbers, the standard uncertainties of the real and
    imaginary components are evaluated, as well as the
    sample correlation coefficient are returned in a
    :obj:`~named_tuples.StandardUncertainty` namedtuple

    Only the values of uncertain numbers are used in calculations. 

    **Example**::

        >>> data = range(15)
        >>> type_a.standard_uncertainty(data)
        1.1547005383792515

        >>> data = [(0.91518731126816899+1.5213442955575518j),
        ... (0.96572684493613492-0.18547192979059401j),
        ... (0.23216598132006649+1.6951311687588568j),
        ... (2.1642786101267397+2.2024333895672563j),
        ... (1.1812532664590505+0.59062101107787357j),
        ... (1.2259264339405165+1.1499373179910186j),
        ... (-0.99422341300318684+1.7359338393131392j),
        ... (1.2122867690240853+0.32535154897909946j),
        ... (2.0122536479379196-0.23283009302603963j),
        ... (1.6770229536619197+0.77195994890476838j)]
        >>> u,r = type_a.standard_uncertainty(data)
        >>> u
        StandardUncertainty(real=0.28881665310241805, imag=0.2655555630050262)
        >>> u.real
        0.28881665310241805
        >>> r
        -0.31374045124595246

    """
    ROOT_N = math.sqrt(len(seq))
    
    if mu is None:
        mu = mean(seq)

    if isinstance(mu,numbers.Real):
        sd = standard_deviation(seq,mu)
        return sd / ROOT_N  
        
    elif isinstance(mu,numbers.Complex):
        sd,r = standard_deviation(seq,mu)
        return StandardUncertainty(sd.real/ROOT_N,sd.imag/ROOT_N),r
        
    else:
        assert False,"unexpected, mu={!r}".format(mu)

#-----------------------------------------------------------------------------------------
def variance_covariance_complex(seq,mu=None):
    """Return the sample variance-covariance matrix

    :arg seq: sequence of data   
    :arg mu: the arithmetic mean of ``seq``

    :returns: a 4-element sequence

    If ``mu`` is ``None`` the mean will be evaluated 
    by :func:`~type_a.mean`.

    ``seq`` may contain numbers or uncertain numbers.
    Only the values of uncertain numbers are used in calculations. 
    
    Variance-covariance matrix elements are returned  
    in a :obj:`~named_tuples.VarianceCovariance` namedtuple; 
    they can be accessed using the 
    attributes ``.rr``, ``.ri``, ``,ir`` and ``.ii``.
        
    **Example**::
    
        >>> data = [(0.91518731126816899+1.5213442955575518j),
        ... (0.96572684493613492-0.18547192979059401j),
        ... (0.23216598132006649+1.6951311687588568j),
        ... (2.1642786101267397+2.2024333895672563j),
        ... (1.1812532664590505+0.59062101107787357j),
        ... (1.2259264339405165+1.1499373179910186j),
        ... (-0.99422341300318684+1.7359338393131392j),
        ... (1.2122867690240853+0.32535154897909946j),
        ... (2.0122536479379196-0.23283009302603963j),
        ... (1.6770229536619197+0.77195994890476838j)]
        >>> type_a.variance_covariance_complex(data)
        VarianceCovariance(rr=0.8341505910928249, ri=-0.24062910264062262, ir=-0.24062910264062262, ii=0.7051975704291644)

        >>> v = type_a.variance_covariance_complex(data)
        >>> v[0]
        0.8341505910928249
        >>> v.rr
        0.8341505910928249
        >>> v.ii
        0.7051975704291644

    """
    N = len(seq)
    df = N-1
    
    zseq = value_seq(seq)
    
    if mu is None:
        mu = mean(zseq)         
        
    mu = complex( mu )
           
    accum_vr = lambda psum,z: psum + (z.real - mu.real)**2
    accum_vi = lambda psum,z: psum + (z.imag - mu.imag)**2
    accum_cv = lambda psum,z: psum + (z.imag - mu.imag)*(z.real - mu.real)
    
    cv_11 = reduce(accum_vr,zseq,0.0) / df 
    cv_22 = reduce(accum_vi,zseq,0.0) / df
    cv_12 = reduce(accum_cv,zseq,0.0) / df

    return VarianceCovariance(cv_11,cv_12,cv_12,cv_22)

#-----------------------------------------------------------------------------------------
def multi_estimate_real(seq_of_seq,labels=None):
    """Return a sequence of uncertain real numbers 

    :arg seq_of_seq: a sequence of sequences of data
    :arg labels: a sequence of `str` labels 
    
    :rtype: seq of :class:`~lib.UncertainReal`

    The sequences in ``seq_of_seq`` must all be the same length.
    Each sequence contains 
    a sample of data associated with a particular quantity. 
    An uncertain number will be created for the quantity
    from sample statistics. The covariance 
    between the different quantities will also be evaluated.
    
    A sequence of elementary uncertain numbers is returned. These uncertain numbers 
    are considered to be related, allowing a degrees-of-freedom calculations 
    to be performed on derived quantities. 

    **Example**::
    
        # From Appendix H2 in the GUM
        
        >>> V = [5.007,4.994,5.005,4.990,4.999]
        >>> I = [19.663E-3,19.639E-3,19.640E-3,19.685E-3,19.678E-3]
        >>> phi = [1.0456,1.0438,1.0468,1.0428,1.0433]
        >>> v,i,p = type_a.multi_estimate_real((V,I,phi),labels=('V','I','phi'))
        >>> v
        ureal(4.999000...,0.0032093613071761...,4, label='V')
        >>> i
        ureal(0.019661,9.471008394041335...e-06,4, label='I')
        >>> p
        ureal(1.044460...,0.0007520638270785...,4, label='phi')
        
        >>> r = v/i*cos(p)
        >>> r
        ureal(127.732169928102...,0.071071407396995...,4.0)
        
    """
    M = len(seq_of_seq)
    N = len(seq_of_seq[0])
    
    if labels is not None and len(labels) != M:
        raise RuntimeError(
            "Incorrect number of labels: '{!r}'".format(labels) 
        ) 
        
    # Calculate the deviations from the mean for each sequence
    means = [ ]
    dev = []
    for i,seq_i in enumerate(seq_of_seq):
        if len(seq_i) != N:
            raise RuntimeError( "{:d}th sequence length inconsistent".format(i) )

        mu_i =  value( sum(seq_i) / N )
        means.append( mu_i )
        dev.append( tuple( value(x_j)-mu_i for x_j in seq_i ) )

    # calculate the covariance matrix
    N_N_1 = N*(N-1)
    u = []
    cv = [] # M elements of len M-1, M-2, ...
    for i,seq_i in enumerate(dev):
        u_i = math.sqrt(
            math.fsum(d_i**2 for d_i in seq_i)/N_N_1
        )
        u.append(u_i)
        cv.append([])
        for seq_j in dev[i+1:]:
            cv[i].append(
                math.fsum(
                    d_i*d_j
                        for d_i,d_j in izip(seq_i,seq_j)
                )/N_N_1
            )

    # Create a list of elementary uncertain numbers
    # to return a list of standard uncertainties
    # to normalise the CV matrix.
    df = N-1
    rtn = []
    for i in xrange(M):
        mu_i = means[i]
        u_i = u[i]
        l_i = labels[i] if labels is not None else ""
        rtn.append( ureal(mu_i,u_i,df,l_i,independent=False) )

    # Create the list of ensemble id's,
    # assign it to the register in the context,
    # set the correlation between nodes
    real_ensemble( rtn, df )
    
    for i in xrange(M):
        u_i = u[i]
        un_i = rtn[i]
        
        for j in xrange(M-1-i):
            cv_ij = cv[i][j]
            if cv_ij != 0.0:
                r =  cv_ij / (u_i*u[i+j+1])
                un_j = rtn[i+j+1]
                set_correlation_real(un_i,un_j,r)

    return rtn

#-----------------------------------------------------------------------------------------
def multi_estimate_complex(seq_of_seq,labels=None):
    """
    Return a sequence of uncertain complex numbers

    :arg seq_of_seq: a sequence of sequences of data
    :arg labels: a sequence of `str` labels
    
    :rtype: a sequence of :class:`~lib.UncertainComplex`
        
    The sequences in ``seq_of_seq`` must all be the same length.
    Each sequence contains data that is associated with 
    a particular quantity. An uncertain number for that quantity will  
    be created from sample statistics. The covariance 
    between the different quantities will also be evaluated.
    
    A sequence of elementary uncertain complex numbers is returned. These   
    uncertain numbers are considered to be related, allowing a degrees-of-freedom  
    calculations to be performed on derived quantities. 
    
    **Example**::
    
        # From Appendix H2 in the GUM
        
        >>> I = [ complex(x) for x in (19.663E-3,19.639E-3,19.640E-3,19.685E-3,19.678E-3) ]
        >>> V = [ complex(x) for x in (5.007,4.994,5.005,4.990,4.999)]
        >>> P = [ complex(0,p) for p in (1.0456,1.0438,1.0468,1.0428,1.0433) ]

        >>> v,i,p = type_a.multi_estimate_complex( (V,I,P) )

        >>> get_correlation(v.real,i.real)
        -0.355311219817512

        >>> z = v/i*exp(p)
        >>> z.real
        ureal(127.732169928102...,0.071071407396995...,4.0)
        >>> get_correlation(z.real,z.imag)
        -0.588429784423515...
        
    """
    M = len(seq_of_seq)
    N = len(seq_of_seq[0])
    
    if labels is not None and len(labels) != M:
        raise RuntimeError( 
            "Incorrect number of labels: '{!r}'".format(labels) 
        ) 

    # 1. Create a 2M sequence of sequences of real values
    x = []
    for i in xrange(M):
        x.append( [ value(z_i.real) for z_i in seq_of_seq[i] ] )
        x.append( [ value(z_i.imag) for z_i in seq_of_seq[i] ] )
        if len(x[-1]) != N:
            raise RuntimeError(
                "{:d}th sequence length is incorrect".format(i)
            )

    TWOM = 2*M
    N_1 = N-1
    N_N_1 = N*N_1

    # 2. Evaluate the means and uncertainties (keep the deviation sequences)
    x_mean = [ value( math.fsum(seq_i) / N ) for seq_i in x ]
    x_u = []
    for i in xrange(TWOM):
        mu_i = x_mean[i]
        x[i] = [ mu_i - x_ij for x_ij in x[i] ]
        x_u.append(
            math.sqrt(
                math.fsum( x_ij**2 for x_ij in x[i] )/N_N_1
            )
        )
    # 3. Define uncertain M complex numbers
    x_influences = []
    rtn = []
    for i in xrange(M):
        j = 2*i
        uc = ucomplex(
            complex(x_mean[j],x_mean[j+1]),
            x_u[j],x_u[j+1],0.0,
            N_1,
            labels[i] if labels is not None else None,
            independent=False
        )
        rtn.append( uc )
        x_influences.extend( (uc.real,uc.imag) )
        

    # 4. Calculate covariances and set correlation coefficients
    for i in xrange(TWOM-1):
        x_i = x[i]
        un_i = x_influences[i]
        for j in xrange(i+1,TWOM):
            x_j = x[j]
            cv = math.fsum( 
                d_i*d_j for d_i,d_j in izip(x_i,x_j)
            )/N_N_1
            if cv != 0.0:
                r = cv/(x_u[i]*x_u[j]) 
                set_correlation_real(un_i,x_influences[j],r)

    complex_ensemble( rtn, N_1 )
    
    return rtn
        
#--------------------------------------------------------------------
def merge(a,b,TOL=1E-13):
    """Combine the uncertainty components of ``a`` and ``b``

    :arg a: an uncertain real or complex number
    :arg b: an uncertain real or complex number
    :arg TOL: float

    :returns:   an uncertain number with the value of ``a`` and the 
                uncertainty components of ``a`` and ``b`` combined

    The absolute difference between the values of ``a`` and ``b`` 
    must be less than ``TOL`` and the components of uncertainty 
    associated with ``a`` and ``b`` must be distinct, otherwise
    a :class:`RuntimeError` will be raised.

    Use this function to combine results from
    type-A and type-B uncertainty analyses 
    performed on a common sequence of data.

    .. note::

        Some judgement will be required as to
        when it is appropriate to merge 
        uncertainty components.

        There is a risk of 'double-counting'
        uncertainty if type-B components
        are contributing to the variability
        observed in the data, and therefore
        assessed in a type-A analysis.

    .. versionchanged:: 1.3.3
       Added the `TOL` keyword argument.

    **Example**::

        # From Appendix H3 in the GUM
        
        # Thermometer readings (degrees C)
        t = (21.521,22.012,22.512,23.003,23.507,
            23.999,24.513,25.002,25.503,26.010,26.511)

        # Observed differences with calibration standard (degrees C)
        b = (-0.171,-0.169,-0.166,-0.159,-0.164,
            -0.165,-0.156,-0.157,-0.159,-0.161,-0.160)

        # Arbitrary offset temperature (degrees C)
        t_0 = 20.0
        
        # Calculate the temperature relative to t_0
        t_rel = [ t_k - t_0 for t_k in t ]

        # A common systematic error in all differences
        e_sys = ureal(0,0.01)
        
        b_type_b = [ b_k + e_sys for b_k in b ]

        # Type-A least-squares regression
        y_1_a, y_2_a = type_a.line_fit(t_rel,b_type_b).a_b

        # Type-B least-squares regression
        y_1_b, y_2_b = type_b.line_fit(t_rel,b_type_b)

        # `y_1` and `y_2` have uncertainty components  
        # related to the type-A analysis as well as the 
        # type-B systematic error
        y_1 = type_a.merge(y_1_a,y_1_b)
        y_2 = type_a.merge(y_2_a,y_2_b)

    """
    if abs( value(a) - value(b) ) > TOL:
        raise RuntimeError(
            "|a - b| = {} > {}: {!r} != {!r}".format(
                abs( value(a) - value(b) ),TOL,a,b
            )
        )
    else:
        return a + (b - value(b))
    
#============================================================================    
if __name__ == "__main__":
    import doctest
    from GTC import *    
    doctest.testmod( optionflags=doctest.NORMALIZE_WHITESPACE )

    



"""
A Python package for evaluating measurement uncertainty
in real and complex quantities.

The method of uncertainty propagation is compatible with the approach described
in the 'Guide to the Expression of Uncertainty in Measurement' - the GUM.

"""
from __future__ import division

import math
import cmath
import sys 

try:
    basestring  # Python 2
    from collections import Sequence
except NameError:
    basestring = (str, bytes)
    from collections.abc import Sequence

#----------------------------------------------------------------------------
# Global constants, etc

# The degrees of freedom is considered infinite above `inf_dof`
inf_dof = 1E5

inf = float('inf')
nan = float('nan')

EPSILON = sys.float_info.epsilon 

# Do not consider strings as sequences
# Note numpy arrays are not detected as sequences
# by this function.
def is_sequence(obj):
    if isinstance(obj, basestring):
        return False
    else:
        return isinstance(obj, Sequence)

#----------------------------------------------------------------------------

__all__ = (
        'ureal'
    ,   'multiple_ureal'
    ,   'multiple_ucomplex'
    ,   'ucomplex'
    ,   'constant'
    ,   'value'
    ,   'uncertainty'
    ,   'variance'
    ,   'dof'
    ,   'label'
    ,   'component'
    ,   'inf'
    ,   'nan'
    ,   'get_correlation'
    ,   'set_correlation'
    ,   'result'
    ,   'get_covariance'
    ,   'cos'
    ,   'sin'
    ,   'tan'
    ,   'acos'
    ,   'asin'
    ,   'atan'
    ,   'atan2'
    ,   'exp'
    ,   'pow'
    ,   'log'
    ,   'log10'
    ,   'sqrt'
    ,   'sinh'
    ,   'cosh'
    ,   'tanh'
    ,   'acosh'
    ,   'asinh'
    ,   'atanh'
    ,   'mag_squared'
    ,   'magnitude'
    ,   'phase'
    ,   'copyright',    'version'
    ,   'reporting',    'rp'
    ,   'function',     'fn'
    ,   'type_b',       'tb'
    ,   'type_a',       'ta'
    ,   'persistence',  'pr'
    ,   'linear_algebra', 'la'
    ,   'math'
    ,   'cmath'
    ,   'inf'
    ,   'nan'
    ,   'EPSILON'
)

#----------------------------------------------------------------------------
version = "1.3.5"

copyright = """Copyright (c) 2021, \
Measurement Standards Laboratory of New Zealand"""

from .core import *



"""
The Vector class is used in the propagation of uncertainty in uncertain numbers 
by forward automatic differentiation. A Vector holds a sequence of components 
of uncertainty
 
The class is like a mapping of key-value pairs. The key is associated with 
the identity of the uncertainty component and the value is the component of
uncertainty.

Propagation of uncertainty is a time-critical process, so elements are 
ordered to help with optimisation. Ordering uses the built-in ``cmp`` 
function applied to the `uid` attribute of a key. 

The function ``is_ordered()`` checks that all reference-value pairs 
in a vector are in order. When ``__debug__`` is ``True``, the order 
of elements is checked when appending (but not during construction).

Copyright (c) 2018, Measurement Standards Laboratory of New Zealand.

"""
try:
    from itertools import izip  # Python 2
except ImportError:
    izip = zip
    cmp = lambda a, b: (a > b) - (a < b)


__all__ = (
    'Vector',
    'scale_vector',
    'scale_vector_twice',
    'merge_vectors',
    'merge_weighted_vectors',
    'merge_weighted_vectors_twice',
    'extend_vector'
)

# The `cmp` function works as follows.
# Tuples and lists are compared lexicographically by comparing 
# corresponding elements. This means that for sequences to compare equal, 
# each element must compare equal, the sequences must be of the same  
# type and have the same length.  
#
# If not equal, the sequences are ordered by comparing the first differing 
# elements. For example, cmp([1,2,x], [1,2,y]) returns the same as cmp(x,y). 
# If the corresponding element does not exist, the shorter sequence is ordered 
# first (for example, [1,2] < [1,2,3]).  
#

# The value of `uid` is a pair of integers used for ordering.
# The end point can be marked by a pair of infinities,
# so any other pair compares less than this marker.
# The `END` object is used in the iteration control structures to
# mark an end point. 
INF = ( float('inf'), float('inf') )
class INF_UID(object): uid = INF

#--------------------------------------------------------------
# TODO:
# Look at the initialisation overhead. It might be better to
# call Vector.copy instead of Vector(copy=v)
#
# Look at using the bisect module to improve performance
#
class Vector(object):
    
    """
    A Vector is a collection of ordered index-value pairs.
    
    """
    
    def __init__(self,**kwargs):
        """
        Vector()                # construct an empty vector
        Vector(copy=v)          # construct a copy of vector 'v'
        Vector(index=i,value=v) # construct a vector with 
                                  values 'v' indexed by 'i'
        
        Index order is preserved when a Vector is constructed,
        but not checked.
        
        """
        if len(kwargs) == 0:
            self._index = []
            self._value = []
            
        elif 'copy' in kwargs:
            copy = kwargs['copy']
            self._index = list(copy._index)
            self._value = list(copy._value)
            
        elif 'index' in kwargs and 'value' in kwargs:
        
            # References to the index and value sequences are stored,
            # instead of making copies. Because of this we need to ensure 
            # the mutability of input sequences 
            idx = kwargs['index']
            if getattr(idx,'append',None) is None:
                raise RuntimeError(
                    "mutable sequence required, got {!r}".format(type(idx))
                )
                
            val = kwargs['value']
            if getattr(val,'append',None) is None:
                raise RuntimeError(
                    "mutable sequence required, got {!r}".format(type(val))
                )
                
            self._index = idx
            self._value = val
            
        elif len(kwargs) != 0:
            raise RuntimeError(
                "unidentified keywords '%s'" % kwargs.keys()
            )

### There appears to be a large overhead associated with this            
##        if __debug__: 
##            # Vector is ordered initially   
##            assert is_ordered(self), self._index
                
    def __len__(self):
        return len(self._index)

    def __iter__(self):
        return iter(self._index)

    def __str__(self):
        return str( dict( izip(self._index,self._value) ) )

    def __contains__(self,i):
        return i in self._index
        
    def __getitem__(self,i):
        return self._value[ self._index.index(i) ]

    def get(self,i,default=None):
        # TODO: use bisect to speed this up, because
        # it impacts on the calculation of covariances
        try:
            return self._value[ self._index.index(i) ]
        except ValueError:
            return default
        
    def items(self):
        return zip(self._index,self._value)
    
    # NB: `keys` and `values` are exposing the 
    # internal structures, which would allow a 
    # client to change, for e.g., an element of 
    # the sequence returned by `keys` and thereby
    # change the underlying Vector. 
    # It would be better to return a copy of the index 
    # and value sequences. 
    # TODO: look at whether there is a significant 
    # time penalty associated with this change.
    def keys(self):
        return self._index

    def values(self):
        return self._value

    def iteritems(self):
        return izip(self._index,self._value)
    
    def iterkeys(self):
        return iter(self._index)

    def itervalues(self):
        return iter(self._value)

    def insert(self,k,v):
        """
        Insert `v` at the correct position for `k` 
        
        """
        assert False, "Not implemented"
        
        # TODO: use bisect to locate the insertion point `i`
        
        self._index.insert(i,k)
        self._value.insert(i,v)
        
    def extend(self,it):
        """
        Append the elements in the iterator `it`
        
        """
        for i,v in it:
            self.append(i,v)
        
    def append(self,i,v):
        """
        Append one (i,v) pair to the vector.

        Note that ordering is not checked unless `__debug__` 
        is `True`. It the responsibility of the client to 
        append elements in the correct order. 
        
        """
        if __debug__:
            # Make sure that order is preserved
            try:
                assert self._index[-1].uid < i.uid, (self._index[-1].uid, i.uid)
            except IndexError: 
                # Don't care about an empty list
                pass
            
        self._index.append(i)
        self._value.append(v)
 
    def pop(self):
        """
        Remove and return the last item 
        
        """
        i = self._index.pop()
        v = self._value.pop()
                
        return i,v 
        
#--------------------------------------------------------------
def is_ordered(v):
    """
    The invariant for all Vector operations.
    
    Returns `True` when the elements are correctly ordered.
    
    """
    try:
        it = v.iterkeys()
        x = next(it)
        while True:
            last, x = x, next(it)
            if x.uid < last.uid: return False
    except StopIteration:
        return True    
            
#--------------------------------------------------------------
def scale_vector(v,w):
    """
    Return a Vector containing the values in `v` multiplied by 'w'

    Parameters
    ----------
    v : Vector
    w : float

    Returns
    -------
    Vector
    
    """
    if w == 1:
        return Vector( copy=v )  
    elif len(v._index):
        return Vector(
            index = v._index,
            value = [ w*x for x in v._value ]
        )
    else:
        # `v` is empty
        return Vector()

        #--------------------------------------------------------------
def merge_vectors(v1,v2):
    """
    Return a vector obtained by merging the vector arguments
    
    Parameters
    ----------
    v1, v2 : Vector

    Returns
    -------
    Vector
    
    """
    if (len(v1) == 0): return v2
    if (len(v2) == 0): return v1
        
    value = []
    index = []

    if v1._index == v2._index:
        # No need to check indices => much faster
        value = [ x1 + x2 for (x1,x2) in izip(v1._value,v2._value) ]
        index = v1._index
    else:
        # A marker at the ends: must remove afterwards!
        v1.append(INF_UID,0)
        v2.append(INF_UID,0)

        # Merge vectors by iteration
        it1 = v1.iteritems()
        it2 = v2.iteritems()

        i1,x1 = next(it1)
        i2,x2 = next(it2)
            
        #------------------------------
        while True:
            
            if i1.uid==INF and i2.uid==INF: break

            case = cmp(i1.uid,i2.uid)
            
            if case == 0:
                # i1 == i2
                index.append(i1)
                value.append( x1+x2 )           
                i1,x1 = next(it1)
                i2,x2 = next(it2)
                    
            elif case < 0:
                # i1 < i2
                index.append(i1)
                value.append( x1 )           
                i1,x1 = next(it1)
            else:
                # i1 > i2
                index.append(i2)
                value.append( x2 )           
                i2,x2 = next(it2)
                
        #------------------------------
        v1.pop() 
        v2.pop()

    return Vector( index=index, value=value )

#--------------------------------------------------------------
def merge_weighted_vectors(v1,w1,v2,w2):
    """
    Return a Vector formed by merging and weighting `v1` and `v2`.

    Parameters
    ----------
    v1, v2 : Vector
    w1, w2 : float
    
    Returns
    -------
    Vector
    
    """
    # If w1==0 or w2==0 the other vector is scaled.
    if (len(v1) == 0) and len(v2):
        return scale_vector(v2,w2)
        
    if (len(v2) == 0) and len(v1):
        return scale_vector(v1,w1)
        
    if (w1 == 1.0) and (w2 == 1.0):
        return merge_vectors(v1,v2)
        
    value = []
    index = []

    if v1._index == v2._index:
        # No need to check indices => much faster
        value = [ w1*x1 + w2*x2 for (x1,x2) in izip(v1._value,v2._value) ]
        index = v1._index
    else:
        # A marker at the ends: must remove afterwards!
        v1.append(INF_UID,0)
        v2.append(INF_UID,0)
        
        # Merge vectors by iteration
        it1 = v1.iteritems()
        it2 = v2.iteritems()

        i1,x1 = next(it1)
        i2,x2 = next(it2)
                    
        #------------------------------
        while True:
            
            if i1.uid==INF and i2.uid==INF: break

            case = cmp(i1.uid,i2.uid)
            
            if case == 0:
                # i1 == i2
                index.append(i1)
                value.append(w1*x1+w2*x2)
                i1,x1 = next(it1)
                i2,x2 = next(it2)
                    
            elif case < 0:
                # i1 < i2
                index.append(i1)
                value.append(w1*x1)
                i1,x1 = next(it1)
            else:
                # i1 > i2
                index.append(i2)
                value.append(w2*x2)
                i2,x2 = next(it2)

        #------------------------------
        v1.pop() 
        v2.pop()
        
    return Vector( value=value, index=index )

#--------------------------------------------------------------
def scale_vector_twice(v,w):
    """
    Return a pair of Vectors weighted by different amounts.
    
    'w' is a pair of weightings.

    Parameters
    ----------
    v : Vector
    w : a pair of weightings

    Returns
    -------
    (Vector,Vector)
    
    """
    w1,w2 = w
    if w1 == 1 and w2 == 1:
        return Vector( copy=v ), Vector( copy=v )
    else:
        return (
            Vector(
                index = v._index,
                value = [ w1*x for x in v._value ]
            ),
            Vector(
                index = v._index,
                value = [ w2*x for x in v._value ]
            )
        )

#--------------------------------------------------------------
def merge_weighted_vectors_twice(v1,w1,v2,w2):
    """
    Return a pair of Vectors obtained by merging `v1` and `v2` 
    with different weightings

    'w1' and 'w2' are weighting pairs (ie, v1 and v2
    are merged two times using different weightings).
    
    Parameters
    ----------
    v1, v2 : Vector
    w1, w2 : weighting pairs

    Returns
    -------
    (Vector,Vector)
    
    """
    if (len(v1) == 0):
        return scale_vector_twice(v2,w2)
        
    if (len(v2) == 0):
        return scale_vector_twice(v1,w1)
        
    value1 = []
    value2 = []
    index = []

    w11,w21 = w1
    w12,w22 = w2
    if v1._index == v2._index:
        # No need to check indices => much faster
        value1 = [ w11*x1 + w12*x2 for (x1,x2) in izip(v1._value,v2._value) ]
        value2 = [ w21*x1 + w22*x2 for (x1,x2) in izip(v1._value,v2._value) ]
        index = v1._index
    else:
        # A marker at the ends: must remove afterwards!
        v1.append(INF_UID,0)
        v2.append(INF_UID,0)

        # Merge vectors by iteration
        it1 = v1.iteritems()
        it2 = v2.iteritems()

        i1,x1 = next(it1)
        i2,x2 = next(it2)
        
        #------------------------------
        while True:
            
            if i1.uid==INF and i2.uid==INF: break

            case = cmp(i1.uid,i2.uid)
            
            if case == 0:
                # i1 == i2
                index.append(i1)
                value1.append(w11*x1+w12*x2)
                value2.append(w21*x1+w22*x2)
                
                i1,x1 = next(it1)
                i2,x2 = next(it2)
                    
            elif case < 0:
                # i1 < i2
                index.append(i1)
                value1.append(w11*x1)
                value2.append(w21*x1)
                i1,x1 = next(it1)

            else:
                # i1 > i2
                index.append(i2)
                value1.append(w12*x2)
                value2.append(w22*x2)
                i2,x2 = next(it2)

        #------------------------------
        v1.pop() 
        v2.pop()

    return (
        Vector( value=value1, index=index )
    ,   Vector( value=value2, index=index )
    )

#--------------------------------------------------------------
def extend_vector(v1,v2):
    """
    Return a Vector containing all the indices in `v1` and `v2`
    
    The values from `v1` are retained, but those from `v2` are 
    not copied; instead those values are zero.

    Parameters
    ----------
    v1, v2 : Vector

    Returns
    -------
    Vector
    
    """
    return merge_weighted_vectors(v1,1.0,v2,0.0)            

    